from Carbon import Evt, Events, Fm, Fonts
from Carbon import Qd, Res, Scrap
from Carbon import TE, TextEdit, Win
from Carbon import App
from Carbon.Appearance import kThemeStateActive, kThemeStateInactive
import waste
import WASTEconst
import Wbase
import Wkeys
import Wcontrols
import PyFontify
import string
from types import TupleType, StringType


class TextBox(Wbase.Widget):

    """A static text widget"""

    def __init__(self, possize, text="", align=TextEdit.teJustLeft,
                            fontsettings=None,
                            backgroundcolor=(0xffff, 0xffff, 0xffff)
                            ):
        if fontsettings is None:
            import W
            fontsettings = W.getdefaultfont()
        Wbase.Widget.__init__(self, possize)
        self.fontsettings = fontsettings
        self.text = text
        self.align = align
        self._backgroundcolor = backgroundcolor

    def draw(self, visRgn = None):
        if self._visible:
            (font, style, size, color) = self.fontsettings
            fontid = GetFNum(font)
            savestate = Qd.GetPenState()
            Qd.TextFont(fontid)
            Qd.TextFace(style)
            Qd.TextSize(size)
            Qd.RGBForeColor(color)
            Qd.RGBBackColor(self._backgroundcolor)
            TE.TETextBox(self.text, self._bounds, self.align)
            Qd.RGBBackColor((0xffff, 0xffff, 0xffff))
            Qd.SetPenState(savestate)

    def get(self):
        return self.text

    def set(self, text):
        self.text = text
        if self._parentwindow and self._parentwindow.wid:
            self.SetPort()
            self.draw()


class _ScrollWidget:

    # to be overridden
    def getscrollrects(self):
        """Return (destrect, viewrect)."""
        return None, None

    # internal method

    def updatescrollbars(self):
        (dl, dt, dr, db), (vl, vt, vr, vb) = self.getscrollrects()
        if self._parent._barx:
            viewwidth = vr - vl
            destwidth = dr - dl
            bar = self._parent._barx
            bar.setmax(destwidth - viewwidth)

            # MacOS 8.1 doesn't automatically disable
            # scrollbars whose max <= min
            bar.enable(destwidth > viewwidth)

            bar.setviewsize(viewwidth)
            bar.set(vl - dl)
        if self._parent._bary:
            viewheight = vb - vt
            destheight = db - dt
            bar = self._parent._bary
            bar.setmax(destheight - viewheight)

            # MacOS 8.1 doesn't automatically disable
            # scrollbars whose max <= min
            bar.enable(destheight > viewheight)

            bar.setviewsize(viewheight)
            bar.set(vt - dt)


UNDOLABELS = [  # Indexed by WEGetUndoInfo() value
        None, "", "typing", "Cut", "Paste", "Clear", "Drag", "Style",
        "Ruler", "backspace", "delete", "transform", "resize"]


class EditText(Wbase.SelectableWidget, _ScrollWidget):

    """A text edit widget, mainly for simple entry fields."""

    def __init__(self, possize, text="",
                            callback=None, inset=(3, 3),
                            fontsettings=None,
                            tabsettings = (32, 0),
                            readonly = 0):
        if fontsettings is None:
            import W
            fontsettings = W.getdefaultfont()
        Wbase.SelectableWidget.__init__(self, possize)
        self.temptext = text
        self.ted = None
        self.selection = None
        self.oldselection = None
        self._callback = callback
        self.changed = 0
        self.selchanged = 0
        self._selected = 0
        self._enabled = 1
        self.wrap = 1
        self.readonly = readonly
        self.fontsettings = fontsettings
        self.tabsettings = tabsettings
        if type(inset) <> TupleType:
            self.inset = (inset, inset)
        else:
            self.inset = inset

    def open(self):
        if not hasattr(self._parent, "_barx"):
            self._parent._barx = None
        if not hasattr(self._parent, "_bary"):
            self._parent._bary = None
        self._calcbounds()
        self.SetPort()
        viewrect, destrect = self._calctextbounds()
        flags = self._getflags()
        self.ted = waste.WENew(destrect, viewrect, flags)
        self.ted.WEInstallTabHooks()
        self.ted.WESetAlignment(WASTEconst.weFlushLeft)
        self.setfontsettings(self.fontsettings)
        self.settabsettings(self.tabsettings)
        self.ted.WEUseText(Res.Resource(self.temptext))
        self.ted.WECalText()
        if self.selection:
            self.setselection(self.selection[0], self.selection[1])
            self.selection = None
        else:
            self.selview()
        self.temptext = None
        self.updatescrollbars()
        self.bind("pageup", self.scrollpageup)
        self.bind("pagedown", self.scrollpagedown)
        self.bind("top", self.scrolltop)
        self.bind("bottom", self.scrollbottom)
        self.selchanged = 0

    def close(self):
        self._parent._barx = None
        self._parent._bary = None
        self.ted = None
        self.temptext = None
        Wbase.SelectableWidget.close(self)

    def textchanged(self, all=0):
        self.changed = 1

    def selectionchanged(self):
        self.selchanged = 1
        self.oldselection = self.getselection()

    def gettabsettings(self):
        return self.tabsettings

    def settabsettings(self, (tabsize, tabmode)):
        self.tabsettings = (tabsize, tabmode)
        if hasattr(self.ted, "WESetTabSize"):
            port = self._parentwindow.wid.GetWindowPort()
            if tabmode:
                (font, style, size, color) = self.getfontsettings()
                savesettings = GetPortFontSettings(port)
                SetPortFontSettings(port, (font, style, size))
                tabsize = Qd.StringWidth(' ' * tabsize)
                SetPortFontSettings(port, savesettings)
            tabsize = max(tabsize, 1)
            self.ted.WESetTabSize(tabsize)
            self.SetPort()
            Qd.EraseRect(self.ted.WEGetViewRect())
            self.ted.WEUpdate(port.visRgn)

    def getfontsettings(self):
        from Carbon import Res
        (font, style, size, color) = self.ted.WEGetRunInfo(0)[4]
        font = Fm.GetFontName(font)
        return (font, style, size, color)

    def setfontsettings(self, (font, style, size, color)):
        self.SetPort()
        if type(font) <> StringType:
            font = Fm.GetFontName(font)
        self.fontsettings = (font, style, size, color)
        fontid = GetFNum(font)
        readonly = self.ted.WEFeatureFlag(WASTEconst.weFReadOnly, -1)
        if readonly:
            self.ted.WEFeatureFlag(WASTEconst.weFReadOnly, 0)
        try:
            self.ted.WEFeatureFlag(WASTEconst.weFInhibitRecal, 1)
            selstart, selend = self.ted.WEGetSelection()
            self.ted.WESetSelection(0, self.ted.WEGetTextLength())
            self.ted.WESetStyle(WASTEconst.weDoFace, (0, 0, 0, (0, 0, 0)))
            self.ted.WESetStyle(WASTEconst.weDoFace |
                                    WASTEconst.weDoColor |
                                    WASTEconst.weDoFont |
                                    WASTEconst.weDoSize,
                                    (fontid, style, size, color))
            self.ted.WEFeatureFlag(WASTEconst.weFInhibitRecal, 0)
            self.ted.WECalText()
            self.ted.WESetSelection(selstart, selend)
        finally:
            if readonly:
                self.ted.WEFeatureFlag(WASTEconst.weFReadOnly, 1)
        viewrect = self.ted.WEGetViewRect()
        Qd.EraseRect(viewrect)
        self.ted.WEUpdate(self._parentwindow.wid.GetWindowPort().visRgn)
        self.selectionchanged()
        self.updatescrollbars()

    def adjust(self, oldbounds):
        self.SetPort()
        # Note: if App.DrawThemeEditTextFrame is ever used, it will be necessary
        # to unconditionally outset the invalidated rectangles, since Appearance
        # frames are drawn outside the bounds.
        if self._selected and self._parentwindow._hasselframes:
            self.GetWindow().InvalWindowRect(Qd.InsetRect(oldbounds, -3, -3))
            self.GetWindow().InvalWindowRect(Qd.InsetRect(self._bounds, -3, -3))
        else:
            self.GetWindow().InvalWindowRect(oldbounds)
            self.GetWindow().InvalWindowRect(self._bounds)
        viewrect, destrect = self._calctextbounds()
        self.ted.WESetViewRect(viewrect)
        self.ted.WESetDestRect(destrect)
        if self.wrap:
            self.ted.WECalText()
        if self.ted.WEGetDestRect()[3] < viewrect[1]:
            self.selview()
        self.updatescrollbars()

    # interface -----------------------
    # selection stuff
    def selview(self):
        self.ted.WESelView()

    def selectall(self):
        self.ted.WESetSelection(0, self.ted.WEGetTextLength())
        self.selectionchanged()
        self.updatescrollbars()

    def selectline(self, lineno, charoffset = 0):
        newselstart, newselend = self.ted.WEGetLineRange(lineno)
        # Autoscroll makes the *end* of the selection visible, which,
        # in the case of a whole line, is the beginning of the *next* line.
        # So sometimes it leaves our line just above the view rect.
        # Let's fool Waste by initially selecting one char less:
        self.ted.WESetSelection(newselstart + charoffset, newselend-1)
        self.ted.WESetSelection(newselstart + charoffset, newselend)
        self.selectionchanged()
        self.updatescrollbars()

    def getselection(self):
        if self.ted:
            return self.ted.WEGetSelection()
        else:
            return self.selection

    def setselection(self, selstart, selend):
        self.selectionchanged()
        if self.ted:
            self.ted.WESetSelection(selstart, selend)
            self.ted.WESelView()
            self.updatescrollbars()
        else:
            self.selection = selstart, selend

    def offsettoline(self, offset):
        return self.ted.WEOffsetToLine(offset)

    def countlines(self):
        return self.ted.WECountLines()

    def getselectedtext(self):
        selstart, selend = self.ted.WEGetSelection()
        return self.ted.WEGetText().data[selstart:selend]

    def expandselection(self):
        oldselstart, oldselend = self.ted.WEGetSelection()
        selstart, selend = min(oldselstart, oldselend), max(oldselstart, oldselend)
        if selstart <> selend and chr(self.ted.WEGetChar(selend-1)) == '\r':
            selend = selend - 1
        newselstart, dummy = self.ted.WEFindLine(selstart, 1)
        dummy, newselend = self.ted.WEFindLine(selend, 1)
        if oldselstart <> newselstart or  oldselend <> newselend:
            self.ted.WESetSelection(newselstart, newselend)
            self.updatescrollbars()
        self.selectionchanged()

    def insert(self, text):
        self.ted.WEInsert(text, None, None)
        self.textchanged()
        self.selectionchanged()

    # text
    def set(self, text):
        if not self.ted:
            self.temptext = text
        else:
            self.ted.WEUseText(Res.Resource(text))
            self.ted.WECalText()
            self.SetPort()
            viewrect, destrect = self._calctextbounds()
            self.ted.WESetViewRect(viewrect)
            self.ted.WESetDestRect(destrect)
            rgn = Qd.NewRgn()
            Qd.RectRgn(rgn, viewrect)
            Qd.EraseRect(viewrect)
            self.draw(rgn)
            self.updatescrollbars()
            self.textchanged(1)

    def get(self):
        if not self._parent:
            return self.temptext
        else:
            return self.ted.WEGetText().data

    # events
    def key(self, char, event):
        (what, message, when, where, modifiers) = event
        if self._enabled and not modifiers & Events.cmdKey or char in Wkeys.arrowkeys:
            self.ted.WEKey(ord(char), modifiers)
            if char not in Wkeys.navigationkeys:
                self.textchanged()
            if char not in Wkeys.scrollkeys:
                self.selectionchanged()
            self.updatescrollbars()
            if self._callback:
                Wbase.CallbackCall(self._callback, 0, char, modifiers)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        self.ted.WEClick(point, modifiers, Evt.TickCount())
        self.selectionchanged()
        self.updatescrollbars()
        return 1

    def idle(self):
        self.SetPort()
        self.ted.WEIdle()

    def rollover(self, point, onoff):
        if onoff:
            Wbase.SetCursor("iBeam")

    def activate(self, onoff):
        self._activated = onoff
        if self._visible:
            self.SetPort()

            # DISABLED!  There are too many places where it is assumed that
            # the frame of an EditText item is 1 pixel, inside the bounds.
            #state = [kThemeStateActive, kThemeStateInactive][not onoff]
            #App.DrawThemeEditTextFrame(Qd.InsetRect(self._bounds, 1, 1), state)

            if self._selected:
                if onoff:
                    self.ted.WEActivate()
                else:
                    self.ted.WEDeactivate()
                self.drawselframe(onoff)

    def select(self, onoff, isclick = 0):
        if Wbase.SelectableWidget.select(self, onoff):
            return
        self.SetPort()
        if onoff:
            self.ted.WEActivate()
            if self._parentwindow._tabbable and not isclick:
                self.selectall()
        else:
            self.ted.WEDeactivate()
        self.drawselframe(onoff)

    def draw(self, visRgn = None):
        if self._visible:
            if not visRgn:
                visRgn = self._parentwindow.wid.GetWindowPort().visRgn
            self.ted.WEUpdate(visRgn)

            # DISABLED!  There are too many places where it is assumed that
            # the frame of an EditText item is 1 pixel, inside the bounds.
            #state = [kThemeStateActive, kThemeStateInactive][not self._activated]
            #App.DrawThemeEditTextFrame(Qd.InsetRect(self._bounds, 1, 1), state)
            Qd.FrameRect(self._bounds)

            if self._selected and self._activated:
                self.drawselframe(1)

    # scrolling
    def scrollpageup(self):
        if self._parent._bary and self._parent._bary._enabled:
            self.vscroll("++")

    def scrollpagedown(self):
        if self._parent._bary and self._parent._bary._enabled:
            self.vscroll("--")

    def scrolltop(self):
        if self._parent._bary and self._parent._bary._enabled:
            self.vscroll(self._parent._bary.getmin())
        if self._parent._barx and self._parent._barx._enabled:
            self.hscroll(self._parent._barx.getmin())

    def scrollbottom(self):
        if self._parent._bary and self._parent._bary._enabled:
            self.vscroll(self._parent._bary.getmax())

    # menu handlers
    def domenu_copy(self, *args):
        selbegin, selend = self.ted.WEGetSelection()
        if selbegin == selend:
            return
        if hasattr(Scrap, 'ZeroScrap'):
            Scrap.ZeroScrap()
        else:
            Scrap.ClearCurrentScrap()
        self.ted.WECopy()
        self.updatescrollbars()

    def domenu_cut(self, *args):
        selbegin, selend = self.ted.WEGetSelection()
        if selbegin == selend:
            return
        if hasattr(Scrap, 'ZeroScrap'):
            Scrap.ZeroScrap()
        else:
            Scrap.ClearCurrentScrap()
        self.ted.WECut()
        self.updatescrollbars()
        self.selview()
        self.textchanged()
        self.selectionchanged()
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, "", None)

    def domenu_paste(self, *args):
        if not self.ted.WECanPaste():
            return
        self.selview()
        self.ted.WEPaste()
        self.updatescrollbars()
        self.textchanged()
        self.selectionchanged()
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, "", None)

    def domenu_clear(self, *args):
        self.ted.WEDelete()
        self.selview()
        self.updatescrollbars()
        self.textchanged()
        self.selectionchanged()
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, "", None)

    def domenu_undo(self, *args):
        which, redo = self.ted.WEGetUndoInfo()
        if not which:
            return
        self.ted.WEUndo()
        self.updatescrollbars()
        self.textchanged()
        self.selectionchanged()
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, "", None)

    def can_undo(self, menuitem):
        #doundo = self.ted.WEFeatureFlag(WASTEconst.weFUndo, -1)
        #print doundo
        #if not doundo:
        #       return 0
        which, redo = self.ted.WEGetUndoInfo()
        if which < len(UNDOLABELS):
            which = UNDOLABELS[which]
        else:
            which = ""
        if which == None:
            return None
        if redo:
            which = "Redo "+which
        else:
            which = "Undo "+which
        menuitem.settext(which)
        return 1

    def domenu_selectall(self, *args):
        self.selectall()

    # private
    def getscrollrects(self):
        return self.ted.WEGetDestRect(), self.ted.WEGetViewRect()

    def vscroll(self, value):
        lineheight = self.ted.WEGetHeight(0, 1)
        dr = self.ted.WEGetDestRect()
        vr = self.ted.WEGetViewRect()
        viewheight = vr[3] - vr[1]
        maxdelta = vr[1] - dr[1]
        mindelta = vr[3] - dr[3]
        if value == "+":
            delta = lineheight
        elif value == "-":
            delta = - lineheight
        elif value == "++":
            delta = viewheight - lineheight
        elif value == "--":
            delta = lineheight - viewheight
        else:   # in thumb
            delta = vr[1] - dr[1] - value
        delta = min(maxdelta, delta)
        delta = max(mindelta, delta)
        delta = int(delta)
        self.ted.WEScroll(0, delta)
        self.updatescrollbars()

    def hscroll(self, value):
        dr = self.ted.WEGetDestRect()
        vr = self.ted.WEGetViewRect()
        destwidth = dr[2] - dr[0]
        viewwidth = vr[2] - vr[0]
        viewoffset = maxdelta = vr[0] - dr[0]
        mindelta = vr[2] - dr[2]
        if value == "+":
            delta = 32
        elif value == "-":
            delta = - 32
        elif value == "++":
            delta = 0.5 * (vr[2] - vr[0])
        elif value == "--":
            delta = 0.5 * (vr[0] - vr[2])
        else:   # in thumb
            delta = vr[0] - dr[0] - value
            #cur = (32767 * viewoffset) / (destwidth - viewwidth)
            #delta = (cur-value)*(destwidth - viewwidth)/32767
            #if abs(delta - viewoffset) <=2:
            #       # compensate for irritating rounding error
            #       delta = viewoffset
        delta = min(maxdelta, delta)
        delta = max(mindelta, delta)
        delta = int(delta)
        self.ted.WEScroll(delta, 0)
        self.updatescrollbars()

    # some internals
    def _getflags(self):
        flags = WASTEconst.weDoAutoScroll | WASTEconst.weDoMonoStyled
        if self.readonly:
            flags = flags | WASTEconst.weDoReadOnly
        else:
            flags = flags | WASTEconst.weDoUndo
        return flags

    def _getviewrect(self):
        return Qd.InsetRect(self._bounds, self.inset[0], self.inset[1])

    def _calctextbounds(self):
        viewrect = l, t, r, b = self._getviewrect()
        if self.ted:
            dl, dt, dr, db = self.ted.WEGetDestRect()
            vl, vt, vr, vb = self.ted.WEGetViewRect()
            yshift = t - vt
            if (db - dt) < (b - t):
                destrect = viewrect
            else:
                destrect = l, dt + yshift, r, db + yshift
        else:
            destrect = viewrect
        return viewrect, destrect


class TextEditor(EditText):

    """A text edit widget."""

    def __init__(self, possize, text="", callback=None, wrap=1, inset=(4, 4),
                            fontsettings=None,
                            tabsettings=(32, 0),
                            readonly=0):
        EditText.__init__(self, possize, text, callback, inset, fontsettings, tabsettings, readonly)
        self.wrap = wrap

    def _getflags(self):
        flags = WASTEconst.weDoAutoScroll | WASTEconst.weDoMonoStyled | \
                        WASTEconst.weDoOutlineHilite
        if self.readonly:
            flags = flags | WASTEconst.weDoReadOnly
        else:
            flags = flags | WASTEconst.weDoUndo
        return flags

    def _getviewrect(self):
        l, t, r, b = self._bounds
        return (l + 5, t + 2, r, b - 2)

    def _calctextbounds(self):
        if self.wrap:
            return EditText._calctextbounds(self)
        else:
            viewrect = l, t, r, b = self._getviewrect()
            if self.ted:
                dl, dt, dr, db = self.ted.WEGetDestRect()
                vl, vt, vr, vb = self.ted.WEGetViewRect()
                xshift = l - vl
                yshift = t - vt
                if (db - dt) < (b - t):
                    yshift = t - dt
                destrect = (dl + xshift, dt + yshift, dr + xshift, db + yshift)
            else:
                destrect = (l, t, r + 5000, b)
            return viewrect, destrect

    def draw(self, visRgn = None):
        if self._visible:
            if not visRgn:
                visRgn = self._parentwindow.wid.GetWindowPort().visRgn
            self.ted.WEUpdate(visRgn)
            if self._selected and self._activated:
                self.drawselframe(1)

    def activate(self, onoff):
        self._activated = onoff
        if self._visible:
            self.SetPort()
            # doesn't draw frame, as EditText.activate does
            if self._selected:
                if onoff:
                    self.ted.WEActivate()
                else:
                    self.ted.WEDeactivate()
                self.drawselframe(onoff)


import re
commentPat = re.compile("[ \t]*(#)")
indentPat = re.compile("[ \t]*")
kStringColor = (0, 0x7fff, 0)
kCommentColor = (0, 0, 0xb000)


class PyEditor(TextEditor):

    """A specialized Python source edit widget"""

    def __init__(self, possize, text="", callback=None, inset=(4, 4),
                            fontsettings=None,
                            tabsettings=(32, 0),
                            readonly=0,
                            debugger=None,
                            file=''):
        TextEditor.__init__(self, possize, text, callback, 0, inset, fontsettings, tabsettings, readonly)
        self.bind("cmd[", self.domenu_shiftleft)
        self.bind("cmd]", self.domenu_shiftright)
        self.bind("cmdshift[", self.domenu_uncomment)
        self.bind("cmdshift]", self.domenu_comment)
        self.bind("cmdshiftd", self.alldirty)
        self.file = file        # only for debugger reference
        self._debugger = debugger
        if debugger:
            debugger.register_editor(self, self.file)
        self._dirty = (0, None)
        self.do_fontify = 0

    #def open(self):
    #       TextEditor.open(self)
    #       if self.do_fontify:
    #               self.fontify()
    #       self._dirty = (None, None)

    def _getflags(self):
        flags = (WASTEconst.weDoDrawOffscreen | WASTEconst.weDoUseTempMem |
                        WASTEconst.weDoAutoScroll | WASTEconst.weDoOutlineHilite)
        if self.readonly:
            flags = flags | WASTEconst.weDoReadOnly
        else:
            flags = flags | WASTEconst.weDoUndo
        return flags

    def textchanged(self, all=0):
        self.changed = 1
        if all:
            self._dirty = (0, None)
            return
        oldsel = self.oldselection
        sel = self.getselection()
        if not sel:
            # XXX what to do?
            return
        selstart, selend = sel
        selstart, selend = min(selstart, selend), max(selstart, selend)
        if oldsel:
            oldselstart, oldselend = min(oldsel), max(oldsel)
            selstart, selend = min(selstart, oldselstart), max(selend, oldselend)
        startline = self.offsettoline(selstart)
        endline = self.offsettoline(selend)
        selstart, _ = self.ted.WEGetLineRange(startline)
        _, selend = self.ted.WEGetLineRange(endline)
        if selstart > 0:
            selstart = selstart - 1
        self._dirty = (selstart, selend)

    def idle(self):
        self.SetPort()
        self.ted.WEIdle()
        if not self.do_fontify:
            return
        start, end = self._dirty
        if start is None:
            return
        textLength = self.ted.WEGetTextLength()
        if end is None:
            end = textLength
        if start >= end:
            self._dirty = (None, None)
        else:
            self.fontify(start, end)
            self._dirty = (None, None)

    def alldirty(self, *args):
        self._dirty = (0, None)

    def fontify(self, start=0, end=None):
        #W.SetCursor('watch')
        if self.readonly:
            self.ted.WEFeatureFlag(WASTEconst.weFReadOnly, 0)
        self.ted.WEFeatureFlag(WASTEconst.weFOutlineHilite, 0)
        self.ted.WEDeactivate()
        self.ted.WEFeatureFlag(WASTEconst.weFAutoScroll, 0)
        self.ted.WEFeatureFlag(WASTEconst.weFUndo, 0)
        pytext = self.get().replace("\r", "\n")
        if end is None:
            end = len(pytext)
        else:
            end = min(end, len(pytext))
        selstart, selend = self.ted.WEGetSelection()
        self.ted.WESetSelection(start, end)
        self.ted.WESetStyle(WASTEconst.weDoFace | WASTEconst.weDoColor,
                        (0, 0, 12, (0, 0, 0)))

        tags = PyFontify.fontify(pytext, start, end)
        styles = {
                'string': (WASTEconst.weDoColor, (0, 0, 0, kStringColor)),
                'keyword': (WASTEconst.weDoFace, (0, 1, 0, (0, 0, 0))),
                'comment': (WASTEconst.weDoFace | WASTEconst.weDoColor, (0, 0, 0, kCommentColor)),
                'identifier': (WASTEconst.weDoColor, (0, 0, 0, (0xbfff, 0, 0)))
        }
        setselection = self.ted.WESetSelection
        setstyle = self.ted.WESetStyle
        for tag, start, end, sublist in tags:
            setselection(start, end)
            mode, style = styles[tag]
            setstyle(mode, style)
        self.ted.WESetSelection(selstart, selend)
        self.SetPort()
        self.ted.WEFeatureFlag(WASTEconst.weFAutoScroll, 1)
        self.ted.WEFeatureFlag(WASTEconst.weFUndo, 1)
        self.ted.WEActivate()
        self.ted.WEFeatureFlag(WASTEconst.weFOutlineHilite, 1)
        if self.readonly:
            self.ted.WEFeatureFlag(WASTEconst.weFReadOnly, 1)

    def domenu_shiftleft(self):
        self.expandselection()
        selstart, selend = self.ted.WEGetSelection()
        selstart, selend = min(selstart, selend), max(selstart, selend)
        snippet = self.getselectedtext()
        lines = string.split(snippet, '\r')
        for i in range(len(lines)):
            if lines[i][:1] == '\t':
                lines[i] = lines[i][1:]
        snippet = string.join(lines, '\r')
        self.insert(snippet)
        self.ted.WESetSelection(selstart, selstart + len(snippet))

    def domenu_shiftright(self):
        self.expandselection()
        selstart, selend = self.ted.WEGetSelection()
        selstart, selend = min(selstart, selend), max(selstart, selend)
        snippet = self.getselectedtext()
        lines = string.split(snippet, '\r')
        for i in range(len(lines) - (not lines[-1])):
            lines[i] = '\t' + lines[i]
        snippet = string.join(lines, '\r')
        self.insert(snippet)
        self.ted.WESetSelection(selstart, selstart + len(snippet))

    def domenu_uncomment(self):
        self.expandselection()
        selstart, selend = self.ted.WEGetSelection()
        selstart, selend = min(selstart, selend), max(selstart, selend)
        snippet = self.getselectedtext()
        lines = string.split(snippet, '\r')
        for i in range(len(lines)):
            m = commentPat.match(lines[i])
            if m:
                pos = m.start(1)
                lines[i] = lines[i][:pos] + lines[i][pos+1:]
        snippet = string.join(lines, '\r')
        self.insert(snippet)
        self.ted.WESetSelection(selstart, selstart + len(snippet))

    def domenu_comment(self):
        self.expandselection()
        selstart, selend = self.ted.WEGetSelection()
        selstart, selend = min(selstart, selend), max(selstart, selend)
        snippet = self.getselectedtext()
        lines = string.split(snippet, '\r')
        indent = 3000 # arbitrary large number...
        for line in lines:
            if string.strip(line):
                m = indentPat.match(line)
                if m:
                    indent = min(indent, m.regs[0][1])
                else:
                    indent = 0
                    break
        for i in range(len(lines) - (not lines[-1])):
            lines[i] = lines[i][:indent] + "#" + lines[i][indent:]
        snippet = string.join(lines, '\r')
        self.insert(snippet)
        self.ted.WESetSelection(selstart, selstart + len(snippet))

    def setfile(self, file):
        self.file = file

    def set(self, text, file = ''):
        oldfile = self.file
        self.file = file
        if self._debugger:
            self._debugger.unregister_editor(self, oldfile)
            self._debugger.register_editor(self, file)
        TextEditor.set(self, text)

    def close(self):
        if self._debugger:
            self._debugger.unregister_editor(self, self.file)
            self._debugger = None
        TextEditor.close(self)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        if self._debugger and self.pt_in_breaks(point):
            self.breakhit(point, modifiers)
        elif self._debugger:
            bl, bt, br, bb = self._getbreakrect()
            Qd.EraseRect((bl, bt, br-1, bb))
            TextEditor.click(self, point, modifiers)
            self.drawbreakpoints()
        else:
            TextEditor.click(self, point, modifiers)
            if self.ted.WEGetClickCount() >= 3:
                # select block with our indent
                lines = string.split(self.get(), '\r')
                selstart, selend = self.ted.WEGetSelection()
                lineno = self.ted.WEOffsetToLine(selstart)
                tabs = 0
                line = lines[lineno]
                while line[tabs:] and line[tabs] == '\t':
                    tabs = tabs + 1
                tabstag = '\t' * tabs
                fromline = 0
                toline = len(lines)
                if tabs:
                    for i in range(lineno - 1, -1, -1):
                        line = lines[i]
                        if line[:tabs] <> tabstag:
                            fromline = i + 1
                            break
                    for i in range(lineno + 1, toline):
                        line = lines[i]
                        if line[:tabs] <> tabstag:
                            toline = i - 1
                            break
                selstart, dummy = self.ted.WEGetLineRange(fromline)
                dummy, selend = self.ted.WEGetLineRange(toline)
                self.ted.WESetSelection(selstart, selend)

    def breakhit(self, point, modifiers):
        if not self.file:
            return
        destrect = self.ted.WEGetDestRect()
        offset, edge = self.ted.WEGetOffset(point)
        lineno = self.ted.WEOffsetToLine(offset) + 1
        if point[1] <= destrect[3]:
            self._debugger.clear_breaks_above(self.file, self.countlines())
            self._debugger.toggle_break(self.file, lineno)
        else:
            self._debugger.clear_breaks_above(self.file, lineno)

    def key(self, char, event):
        (what, message, when, where, modifiers) = event
        if modifiers & Events.cmdKey and not char in Wkeys.arrowkeys:
            return
        if char == '\r':
            selstart, selend = self.ted.WEGetSelection()
            selstart, selend = min(selstart, selend), max(selstart, selend)
            lastchar = chr(self.ted.WEGetChar(selstart-1))
            if lastchar <> '\r' and selstart:
                pos, dummy = self.ted.WEFindLine(selstart, 0)
                lineres = Res.Resource('')
                self.ted.WECopyRange(pos, selstart, lineres, None, None)
                line = lineres.data + '\n'
                tabcount = self.extratabs(line)
                self.ted.WEKey(ord('\r'), 0)
                for i in range(tabcount):
                    self.ted.WEKey(ord('\t'), 0)
            else:
                self.ted.WEKey(ord('\r'), 0)
        elif char in ')]}':
            self.ted.WEKey(ord(char), modifiers)
            self.balanceparens(char)
        else:
            self.ted.WEKey(ord(char), modifiers)
        if char not in Wkeys.navigationkeys:
            self.textchanged()
        self.selectionchanged()
        self.updatescrollbars()

    def balanceparens(self, char):
        if char == ')':
            target = '('
        elif char == ']':
            target = '['
        elif char == '}':
            target = '{'
        recursionlevel = 1
        selstart, selend = self.ted.WEGetSelection()
        count = min(selstart, selend) - 2
        mincount = max(0, count - 2048)
        lastquote = None
        while count > mincount:
            testchar = chr(self.ted.WEGetChar(count))
            if testchar in "\"'" and chr(self.ted.WEGetChar(count - 1)) <> '\\':
                if lastquote == testchar:
                    recursionlevel = recursionlevel - 1
                    lastquote = None
                elif not lastquote:
                    recursionlevel = recursionlevel + 1
                    lastquote = testchar
            elif not lastquote and testchar == char:
                recursionlevel = recursionlevel + 1
            elif not lastquote and testchar == target:
                recursionlevel = recursionlevel - 1
                if recursionlevel == 0:
                    import time
                    autoscroll = self.ted.WEFeatureFlag(WASTEconst.weFAutoScroll, -1)
                    if autoscroll:
                        self.ted.WEFeatureFlag(WASTEconst.weFAutoScroll, 0)
                    self.ted.WESetSelection(count, count + 1)
                    self._parentwindow.wid.GetWindowPort().QDFlushPortBuffer(None)  # needed under OSX
                    time.sleep(0.2)
                    self.ted.WESetSelection(selstart, selend)
                    if autoscroll:
                        self.ted.WEFeatureFlag(WASTEconst.weFAutoScroll, 1)
                    break
            count = count - 1

    def extratabs(self, line):
        tabcount = 0
        for c in line:
            if c <> '\t':
                break
            tabcount = tabcount + 1
        last = 0
        cleanline = ''
        tags = PyFontify.fontify(line)
        # strip comments and strings
        for tag, start, end, sublist in tags:
            if tag in ('string', 'comment'):
                cleanline = cleanline + line[last:start]
                last = end
        cleanline = cleanline + line[last:]
        cleanline = string.strip(cleanline)
        if cleanline and cleanline[-1] == ':':
            tabcount = tabcount + 1
        else:
            # extra indent after unbalanced (, [ or {
            for open, close in (('(', ')'), ('[', ']'), ('{', '}')):
                count = string.count(cleanline, open)
                if count and count > string.count(cleanline, close):
                    tabcount = tabcount + 2
                    break
        return tabcount

    def rollover(self, point, onoff):
        if onoff:
            if self._debugger and self.pt_in_breaks(point):
                Wbase.SetCursor("arrow")
            else:
                Wbase.SetCursor("iBeam")

    def draw(self, visRgn = None):
        TextEditor.draw(self, visRgn)
        if self._debugger:
            self.drawbreakpoints()

    def showbreakpoints(self, onoff):
        if (not not self._debugger) <> onoff:
            if onoff:
                if not __debug__:
                    import W
                    raise W.AlertError, "Can't debug in \"Optimize bytecode\" mode.\r(see \"Default startup options\" in EditPythonPreferences)"
                import PyDebugger
                self._debugger = PyDebugger.getdebugger()
                self._debugger.register_editor(self, self.file)
            elif self._debugger:
                self._debugger.unregister_editor(self, self.file)
                self._debugger = None
            self.adjust(self._bounds)

    def togglebreakpoints(self):
        self.showbreakpoints(not self._debugger)

    def clearbreakpoints(self):
        if self.file:
            self._debugger.clear_all_file_breaks(self.file)

    def editbreakpoints(self):
        if self._debugger:
            self._debugger.edit_breaks()
            self._debugger.breaksviewer.selectfile(self.file)

    def drawbreakpoints(self, eraseall = 0):
        breakrect = bl, bt, br, bb = self._getbreakrect()
        br = br - 1
        self.SetPort()
        Qd.PenPat(Qd.GetQDGlobalsGray())
        Qd.PaintRect((br, bt, br + 1, bb))
        Qd.PenNormal()
        self._parentwindow.tempcliprect(breakrect)
        Qd.RGBForeColor((0xffff, 0, 0))
        try:
            lasttop = bt
            self_ted = self.ted
            Qd_PaintOval = Qd.PaintOval
            Qd_EraseRect = Qd.EraseRect
            for lineno in self._debugger.get_file_breaks(self.file):
                start, end = self_ted.WEGetLineRange(lineno - 1)
                if lineno <> self_ted.WEOffsetToLine(start) + 1:
                    # breakpoints beyond our text: erase rest, and back out
                    Qd_EraseRect((bl, lasttop, br, bb))
                    break
                (x, y), h = self_ted.WEGetPoint(start, 0)
                bottom = y + h
                #print y, (lasttop, bottom)
                if bottom > lasttop:
                    Qd_EraseRect((bl, lasttop, br, y + h * eraseall))
                    lasttop = bottom
                redbullet = bl + 2, y + 3, bl + 8, y + 9
                Qd_PaintOval(redbullet)
            else:
                Qd_EraseRect((bl, lasttop, br, bb))
            Qd.RGBForeColor((0, 0, 0))
        finally:
            self._parentwindow.restoreclip()

    def updatescrollbars(self):
        if self._debugger:
            self.drawbreakpoints(1)
        TextEditor.updatescrollbars(self)

    def pt_in_breaks(self, point):
        return Qd.PtInRect(point, self._getbreakrect())

    def _getbreakrect(self):
        if self._debugger:
            l, t, r, b = self._bounds
            return (l+1, t+1, l + 12, b-1)
        else:
            return (0, 0, 0, 0)

    def _getviewrect(self):
        l, t, r, b = self._bounds
        if self._debugger:
            return (l + 17, t + 2, r, b - 2)
        else:
            return (l + 5, t + 2, r, b - 2)

    def _calctextbounds(self):
        viewrect = l, t, r, b = self._getviewrect()
        if self.ted:
            dl, dt, dr, db = self.ted.WEGetDestRect()
            vl, vt, vr, vb = self.ted.WEGetViewRect()
            xshift = l - vl
            yshift = t - vt
            if (db - dt) < (b - t):
                yshift = t - dt
            destrect = (dl + xshift, dt + yshift, dr + xshift, db + yshift)
        else:
            destrect = (l, t, r + 5000, b)
        return viewrect, destrect


def GetFNum(fontname):
    """Same as Fm.GetFNum(), but maps a missing font to Monaco instead of the system font."""
    if fontname <> Fm.GetFontName(0):
        fontid = Fm.GetFNum(fontname)
        if fontid == 0:
            fontid = Fonts.monaco
    else:
        fontid = 0
    return fontid

# b/w compat. Anyone using this?
GetFName = Fm.GetFontName

def GetPortFontSettings(port):
    return Fm.GetFontName(port.GetPortTextFont()), port.GetPortTextFace(), port.GetPortTextSize()

def SetPortFontSettings(port, (font, face, size)):
    saveport = Qd.GetPort()
    Qd.SetPort(port)
    Qd.TextFont(GetFNum(font))
    Qd.TextFace(face)
    Qd.TextSize(size)
    Qd.SetPort(saveport)



# Python MSI Generator
# (C) 2003 Martin v. Loewis
# See "FOO" in comments refers to MSDN sections with the title FOO.
import msilib, schema, sequence, os, glob, time
from msilib import Feature, CAB, Directory, Dialog, Binary, add_data
import uisample
from win32com.client import constants

# Settings can be overridden in config.py below
# 1 for Itanium build
msilib.Win64 = 0
# 0 for official python.org releases
# 1 for intermediate releases by anybody, with
# a new product code for every package.
snapshot = 1
# 1 means that file extension is px, not py,
# and binaries start with x
testpackage = 0
# Location of build tree
srcdir = os.path.abspath("../..")
# Text to be displayed as the version in dialogs etc.
# goes into file name and ProductCode. Defaults to
# current_version.day for Snapshot, current_version otherwise
full_current_version = None
# Is Tcl available at all?
have_tcl = True

try:
    from config import *
except ImportError:
    pass

# Extract current version from Include/patchlevel.h
lines = open(srcdir + "/Include/patchlevel.h").readlines()
major = minor = micro = level = serial = None
levels = {
    'PY_RELEASE_LEVEL_ALPHA':0xA,
    'PY_RELEASE_LEVEL_BETA': 0xB,
    'PY_RELEASE_LEVEL_GAMMA':0xC,
    'PY_RELEASE_LEVEL_FINAL':0xF
    }
for l in lines:
    if not l.startswith("#define"):
        continue
    l = l.split()
    if len(l) != 3:
        continue
    _, name, value = l
    if name == 'PY_MAJOR_VERSION': major = value
    if name == 'PY_MINOR_VERSION': minor = value
    if name == 'PY_MICRO_VERSION': micro = value
    if name == 'PY_RELEASE_LEVEL': level = levels[value]
    if name == 'PY_RELEASE_SERIAL': serial = value

short_version = major+"."+minor
# See PC/make_versioninfo.c
FIELD3 = 1000*int(micro) + 10*level + int(serial)
current_version = "%s.%d" % (short_version, FIELD3)

# This should never change. The UpgradeCode of this package can be
# used in the Upgrade table of future packages to make the future
# package replace this one. See "UpgradeCode Property".
upgrade_code_snapshot='{92A24481-3ECB-40FC-8836-04B7966EC0D5}'
upgrade_code='{65E6DE48-A358-434D-AA4F-4AF72DB4718F}'

# This should be extended for each Python release.
# The product code must change whenever the name of the MSI file
# changes, and when new component codes are issued for existing
# components. See "Changing the Product Code". As we change the
# component codes with every build, we need a new product code
# each time. For intermediate (snapshot) releases, they are automatically
# generated. For official releases, we record the product codes,
# so people can refer to them.
product_codes = {
    '2.4.101': '{0e9b4d8e-6cda-446e-a208-7b92f3ddffa0}', # 2.4a1, released as a snapshot
    '2.4.102': '{1b998745-4901-4edb-bc52-213689e1b922}', # 2.4a2
    '2.4.103': '{33fc8bd2-1e8f-4add-a40a-ade2728d5942}', # 2.4a3
    '2.4.111': '{51a7e2a8-2025-4ef0-86ff-e6aab742d1fa}', # 2.4b1
    '2.4.112': '{4a5e7c1d-c659-4fe3-b8c9-7c65bd9c95a5}', # 2.4b2
    '2.4.121': '{75508821-a8e9-40a8-95bd-dbe6033ddbea}', # 2.4c1
    '2.4.122': '{83a9118b-4bdd-473b-afc3-bcb142feca9e}', # 2.4c2
    '2.4.150': '{82d9302e-f209-4805-b548-52087047483a}', # 2.4.0
}

if snapshot:
    current_version = "%s.%s.%s" % (major, minor, int(time.time()/3600/24))
    product_code = msilib.gen_uuid()
else:
    product_code = product_codes[current_version]

if full_current_version is None:
    full_current_version = current_version

extensions = [
    'bz2.pyd',
    'pyexpat.pyd',
    'select.pyd',
    'unicodedata.pyd',
    'winsound.pyd',
    'zlib.pyd',
    '_bsddb.pyd',
    '_socket.pyd',
    '_ssl.pyd',
    '_testcapi.pyd',
    '_tkinter.pyd',
]

if major+minor <= "23":
    extensions.extend([
    '_csv.pyd',
    '_sre.pyd',
    '_symtable.pyd',
    '_winreg.pyd',
    'datetime.pyd'
    'mmap.pyd',
    'parser.pyd',
    ])

if testpackage:
    ext = 'px'
    testprefix = 'x'
else:
    ext = 'py'
    testprefix = ''

if msilib.Win64:
    SystemFolderName = "[SystemFolder64]"
else:
    SystemFolderName = "[SystemFolder]"

msilib.reset()

# condition in which to install pythonxy.dll in system32:
# a) it is Windows 9x or
# b) it is NT, the user is privileged, and has chosen per-machine installation
sys32cond = "(Windows9x or (Privileged and ALLUSERS))"

# Install extensions if the feature goes into
# INSTALLSTATE_ADVERTISED or INSTALLSTATE_LOCAL
want_extensions = "(&Extensions = 1) or (&Extensions = 3)"

def build_database():
    """Generate an empty database, with just the schema and the
    Summary information stream."""
    if snapshot:
        uc = upgrade_code_snapshot
    else:
        uc = upgrade_code
    # schema represents the installer 2.0 database schema.
    # sequence is the set of standard sequences
    # (ui/execute, admin/advt/install)
    if msilib.Win64:
        w64 = ".ia64"
    else:
        w64 = ""
    db = msilib.init_database("python-%s%s.msi" % (full_current_version, w64),
                  schema, ProductName="Python "+full_current_version,
                  ProductCode=product_code,
                  ProductVersion=current_version,
                  Manufacturer=u"Martin v. L\xf6wis")
    # The default sequencing of the RemoveExistingProducts action causes
    # removal of files that got just installed. Place it after
    # InstallInitialize, so we first uninstall everything, but still roll
    # back in case the installation is interrupted
    msilib.change_sequence(sequence.InstallExecuteSequence,
                           "RemoveExistingProducts", 1510)
    # Conditionalize Extension information
    msilib.change_sequence(sequence.InstallExecuteSequence,
                            "RegisterClassInfo", cond=want_extensions)
    msilib.change_sequence(sequence.InstallExecuteSequence,
                            "RegisterExtensionInfo", cond=want_extensions)
    msilib.change_sequence(sequence.InstallExecuteSequence,
                            "RegisterProgIdInfo", cond=want_extensions)
    msilib.change_sequence(sequence.InstallExecuteSequence,
                            "RegisterMIMEInfo", cond=want_extensions)
    msilib.add_tables(db, sequence)
    # We cannot set ALLUSERS in the property table, as this cannot be
    # reset if the user choses a per-user installation. Instead, we
    # maintain WhichUsers, which can be "ALL" or "JUSTME". The UI manages
    # this property, and when the execution starts, ALLUSERS is set
    # accordingly.
    add_data(db, "Property", [("UpgradeCode", uc),
                              ("WhichUsers", "ALL"),
                             ])
    db.Commit()
    return db

def remove_old_versions(db):
    "Fill the upgrade table."
    start = "%s.%s.0" % (major, minor)
    # This requests that feature selection states of an older
    # installation should be forwarded into this one. Upgrading
    # requires that both the old and the new installation are
    # either both per-machine or per-user.
    migrate_features = 1
    # See "Upgrade Table". We remove releases with the same major and
    # minor version. For an snapshot, we remove all earlier snapshots. For
    # a release, we remove all snapshots, and all earlier releases.
    if snapshot:
        add_data(db, "Upgrade",
            [(upgrade_code_snapshot, start,
              current_version,
              None,                     # Ignore language
              migrate_features,
              None,                     # Migrate ALL features
              "REMOVEOLDSNAPSHOT")])
        props = "REMOVEOLDSNAPSHOT"
    else:
        add_data(db, "Upgrade",
            [(upgrade_code, start, current_version,
              None, migrate_features, None, "REMOVEOLDVERSION"),
             (upgrade_code_snapshot, start, "%s.%d.0" % (major, int(minor)+1),
              None, migrate_features, None, "REMOVEOLDSNAPSHOT")])
        props = "REMOVEOLDSNAPSHOT;REMOVEOLDVERSION"
    # Installer collects the product codes of the earlier releases in
    # these properties. In order to allow modification of the properties,
    # they must be declared as secure. See "SecureCustomProperties Property"
    add_data(db, "Property", [("SecureCustomProperties", props)])

class PyDialog(Dialog):
    """Dialog class with a fixed layout: controls at the top, then a ruler,
    then a list of buttons: back, next, cancel. Optionally a bitmap at the
    left."""
    def __init__(self, *args, **kw):
        """Dialog(database, name, x, y, w, h, attributes, title, first,
        default, cancel, bitmap=true)"""
        Dialog.__init__(self, *args)
        ruler = self.h - 36
        bmwidth = 152*ruler/328
        if kw.get("bitmap", True):
            self.bitmap("Bitmap", 0, 0, bmwidth, ruler, "PythonWin")
        self.line("BottomLine", 0, ruler, self.w, 0)

    def title(self, title):
        "Set the title text of the dialog at the top."
        # name, x, y, w, h, flags=Visible|Enabled|Transparent|NoPrefix,
        # text, in VerdanaBold10
        self.text("Title", 135, 10, 220, 60, 0x30003,
                  r"{\VerdanaBold10}%s" % title)

    def back(self, title, next, name = "Back", active = 1):
        """Add a back button with a given title, the tab-next button,
        its name in the Control table, possibly initially disabled.

        Return the button, so that events can be associated"""
        if active:
            flags = 3 # Visible|Enabled
        else:
            flags = 1 # Visible
        return self.pushbutton(name, 180, self.h-27 , 56, 17, flags, title, next)

    def cancel(self, title, next, name = "Cancel", active = 1):
        """Add a cancel button with a given title, the tab-next button,
        its name in the Control table, possibly initially disabled.

        Return the button, so that events can be associated"""
        if active:
            flags = 3 # Visible|Enabled
        else:
            flags = 1 # Visible
        return self.pushbutton(name, 304, self.h-27, 56, 17, flags, title, next)

    def next(self, title, next, name = "Next", active = 1):
        """Add a Next button with a given title, the tab-next button,
        its name in the Control table, possibly initially disabled.

        Return the button, so that events can be associated"""
        if active:
            flags = 3 # Visible|Enabled
        else:
            flags = 1 # Visible
        return self.pushbutton(name, 236, self.h-27, 56, 17, flags, title, next)

    def xbutton(self, name, title, next, xpos):
        """Add a button with a given title, the tab-next button,
        its name in the Control table, giving its x position; the
        y-position is aligned with the other buttons.

        Return the button, so that events can be associated"""
        return self.pushbutton(name, int(self.w*xpos - 28), self.h-27, 56, 17, 3, title, next)

def add_ui(db):
    x = y = 50
    w = 370
    h = 300
    title = "[ProductName] Setup"

    # see "Dialog Style Bits"
    modal = 3      # visible | modal
    modeless = 1   # visible
    track_disk_space = 32

    add_data(db, 'ActionText', uisample.ActionText)
    add_data(db, 'UIText', uisample.UIText)

    # Bitmaps
    if not os.path.exists(srcdir+r"\PC\python_icon.exe"):
        raise "Run icons.mak in PC directory"
    add_data(db, "Binary",
             [("PythonWin", msilib.Binary(srcdir+r"\PCbuild\installer.bmp")), # 152x328 pixels
              ("py.ico",msilib.Binary(srcdir+r"\PC\py.ico")),
             ])
    add_data(db, "Icon",
             [("python_icon.exe", msilib.Binary(srcdir+r"\PC\python_icon.exe"))])

    # Scripts
    open("inst.vbs","w").write("""
    Function CheckDir()
      Set FSO = CreateObject("Scripting.FileSystemObject")
      if FSO.FolderExists(Session.Property("TARGETDIR")) then
        Session.Property("TargetExists") = "1"
      else
        Session.Property("TargetExists") = "0"
      end if
    End Function
    """)
    add_data(db, "Binary", [("Script", msilib.Binary("inst.vbs"))])
    # See "Custom Action Type 6"
    add_data(db, "CustomAction", [("CheckDir", 6, "Script", "CheckDir")])
    os.unlink("inst.vbs")


    # UI customization properties
    add_data(db, "Property",
             # See "DefaultUIFont Property"
             [("DefaultUIFont", "DlgFont8"),
              # See "ErrorDialog Style Bit"
              ("ErrorDialog", "ErrorDlg"),
              ("Progress1", "Install"),   # modified in maintenance type dlg
              ("Progress2", "installs"),
              ("MaintenanceForm_Action", "Repair")])

    # Fonts, see "TextStyle Table"
    add_data(db, "TextStyle",
             [("DlgFont8", "Tahoma", 9, None, 0),
              ("DlgFontBold8", "Tahoma", 8, None, 1), #bold
              ("VerdanaBold10", "Verdana", 10, None, 1),
             ])

    # See "CustomAction Table"
    add_data(db, "CustomAction", [
        # msidbCustomActionTypeFirstSequence + msidbCustomActionTypeTextData + msidbCustomActionTypeProperty
        # See "Custom Action Type 51",
        # "Custom Action Execution Scheduling Options"
        ("InitialTargetDir", 307, "TARGETDIR",
         "[WindowsVolume]Python%s%s" % (major, minor)),
        ("SetDLLDirToTarget", 307, "DLLDIR", "[TARGETDIR]"),
        ("SetDLLDirToSystem32", 307, "DLLDIR", SystemFolderName),
        # msidbCustomActionTypeExe + msidbCustomActionTypeSourceFile
        # See "Custom Action Type 18"
        ("CompilePyc", 18, "python.exe", r"[TARGETDIR]Lib\compileall.py [TARGETDIR]Lib"),
        ("CompilePyo", 18, "python.exe", r"-O [TARGETDIR]Lib\compileall.py [TARGETDIR]Lib")
        ])

    # UI Sequences, see "InstallUISequence Table", "Using a Sequence Table"
    # Numbers indicate sequence; see sequence.py for how these action integrate
    add_data(db, "InstallUISequence",
             [("PrepareDlg", "Not Privileged or Windows9x or Installed", 140),
              ("WhichUsersDlg", "Privileged and not Windows9x and not Installed", 141),
              ("InitialTargetDir", 'TARGETDIR=""', 750),
              # In the user interface, assume all-users installation if privileged.
              ("SetDLLDirToSystem32", 'DLLDIR="" and ' + sys32cond, 751),
              ("SetDLLDirToTarget", 'DLLDIR="" and not ' + sys32cond, 752),
              ("SelectDirectoryDlg", "Not Installed", 1230),
              # XXX no support for resume installations yet
              #("ResumeDlg", "Installed AND (RESUME OR Preselected)", 1240),
              ("MaintenanceTypeDlg", "Installed AND NOT RESUME AND NOT Preselected", 1250),
              ("ProgressDlg", None, 1280)])
    add_data(db, "AdminUISequence",
             [("InitialTargetDir", 'TARGETDIR=""', 750),
              ("SetDLLDirToTarget", 'DLLDIR=""', 751),
             ])

    # Execute Sequences
    add_data(db, "InstallExecuteSequence",
            [("InitialTargetDir", 'TARGETDIR=""', 750),
             ("SetDLLDirToSystem32", 'DLLDIR="" and ' + sys32cond, 751),
             ("SetDLLDirToTarget", 'DLLDIR="" and not ' + sys32cond, 752),
             ("CompilePyc", "COMPILEALL", 6800),
             ("CompilePyo", "COMPILEALL", 6801),
            ])
    add_data(db, "AdminExecuteSequence",
            [("InitialTargetDir", 'TARGETDIR=""', 750),
             ("SetDLLDirToTarget", 'DLLDIR=""', 751),
             ("CompilePyc", "COMPILEALL", 6800),
             ("CompilePyo", "COMPILEALL", 6801),
            ])

    #####################################################################
    # Standard dialogs: FatalError, UserExit, ExitDialog
    fatal=PyDialog(db, "FatalError", x, y, w, h, modal, title,
                 "Finish", "Finish", "Finish")
    fatal.title("[ProductName] Installer ended prematurely")
    fatal.back("< Back", "Finish", active = 0)
    fatal.cancel("Cancel", "Back", active = 0)
    fatal.text("Description1", 135, 70, 220, 80, 0x30003,
               "[ProductName] setup ended prematurely because of an error.  Your system has not been modified.  To install this program at a later time, please run the installation again.")
    fatal.text("Description2", 135, 155, 220, 20, 0x30003,
               "Click the Finish button to exit the Installer.")
    c=fatal.next("Finish", "Cancel", name="Finish")
    # See "ControlEvent Table". Parameters are the event, the parameter
    # to the action, and optionally the condition for the event, and the order
    # of events.
    c.event("EndDialog", "Exit")

    user_exit=PyDialog(db, "UserExit", x, y, w, h, modal, title,
                 "Finish", "Finish", "Finish")
    user_exit.title("[ProductName] Installer was interrupted")
    user_exit.back("< Back", "Finish", active = 0)
    user_exit.cancel("Cancel", "Back", active = 0)
    user_exit.text("Description1", 135, 70, 220, 80, 0x30003,
               "[ProductName] setup was interrupted.  Your system has not been modified.  "
               "To install this program at a later time, please run the installation again.")
    user_exit.text("Description2", 135, 155, 220, 20, 0x30003,
               "Click the Finish button to exit the Installer.")
    c = user_exit.next("Finish", "Cancel", name="Finish")
    c.event("EndDialog", "Exit")

    exit_dialog = PyDialog(db, "ExitDialog", x, y, w, h, modal, title,
                         "Finish", "Finish", "Finish")
    exit_dialog.title("Completing the [ProductName] Installer")
    exit_dialog.back("< Back", "Finish", active = 0)
    exit_dialog.cancel("Cancel", "Back", active = 0)
    exit_dialog.text("Acknowledgements", 135, 95, 220, 120, 0x30003,
      "Special Windows thanks to:\n"
      "    LettError, Erik van Blokland, for the \n"
      "    Python for Windows graphic.\n"
      "       http://www.letterror.com/\n"
      "\n"
      "    Mark Hammond, without whose years of freely \n"
      "    shared Windows expertise, Python for Windows \n"
      "    would still be Python for DOS.")

    exit_dialog.text("Description", 135, 235, 220, 20, 0x30003,
               "Click the Finish button to exit the Installer.")
    c = exit_dialog.next("Finish", "Cancel", name="Finish")
    c.event("EndDialog", "Return")

    #####################################################################
    # Required dialog: FilesInUse, ErrorDlg
    inuse = PyDialog(db, "FilesInUse",
                     x, y, w, h,
                     19,                # KeepModeless|Modal|Visible
                     title,
                     "Retry", "Retry", "Retry", bitmap=False)
    inuse.text("Title", 15, 6, 200, 15, 0x30003,
               r"{\DlgFontBold8}Files in Use")
    inuse.text("Description", 20, 23, 280, 20, 0x30003,
               "Some files that need to be updated are currently in use.")
    inuse.text("Text", 20, 55, 330, 50, 3,
               "The following applications are using files that need to be updated by this setup. Close these applications and then click Retry to continue the installation or Cancel to exit it.")
    inuse.control("List", "ListBox", 20, 107, 330, 130, 7, "FileInUseProcess",
                  None, None, None)
    c=inuse.back("Exit", "Ignore", name="Exit")
    c.event("EndDialog", "Exit")
    c=inuse.next("Ignore", "Retry", name="Ignore")
    c.event("EndDialog", "Ignore")
    c=inuse.cancel("Retry", "Exit", name="Retry")
    c.event("EndDialog","Retry")


    # See "Error Dialog". See "ICE20" for the required names of the controls.
    error = Dialog(db, "ErrorDlg",
                   50, 10, 330, 101,
                   65543,       # Error|Minimize|Modal|Visible
                   title,
                   "ErrorText", None, None)
    error.text("ErrorText", 50,9,280,48,3, "")
    error.control("ErrorIcon", "Icon", 15, 9, 24, 24, 5242881, None, "py.ico", None, None)
    error.pushbutton("N",120,72,81,21,3,"No",None).event("EndDialog","ErrorNo")
    error.pushbutton("Y",240,72,81,21,3,"Yes",None).event("EndDialog","ErrorYes")
    error.pushbutton("A",0,72,81,21,3,"Abort",None).event("EndDialog","ErrorAbort")
    error.pushbutton("C",42,72,81,21,3,"Cancel",None).event("EndDialog","ErrorCancel")
    error.pushbutton("I",81,72,81,21,3,"Ignore",None).event("EndDialog","ErrorIgnore")
    error.pushbutton("O",159,72,81,21,3,"Ok",None).event("EndDialog","ErrorOk")
    error.pushbutton("R",198,72,81,21,3,"Retry",None).event("EndDialog","ErrorRetry")

    #####################################################################
    # Global "Query Cancel" dialog
    cancel = Dialog(db, "CancelDlg", 50, 10, 260, 85, 3, title,
                    "No", "No", "No")
    cancel.text("Text", 48, 15, 194, 30, 3,
                "Are you sure you want to cancel [ProductName] installation?")
    cancel.control("Icon", "Icon", 15, 15, 24, 24, 5242881, None,
                   "py.ico", None, None)
    c=cancel.pushbutton("Yes", 72, 57, 56, 17, 3, "Yes", "No")
    c.event("EndDialog", "Exit")

    c=cancel.pushbutton("No", 132, 57, 56, 17, 3, "No", "Yes")
    c.event("EndDialog", "Return")

    #####################################################################
    # Global "Wait for costing" dialog
    costing = Dialog(db, "WaitForCostingDlg", 50, 10, 260, 85, modal, title,
                     "Return", "Return", "Return")
    costing.text("Text", 48, 15, 194, 30, 3,
                 "Please wait while the installer finishes determining your disk space requirements.")
    costing.control("Icon", "Icon", 15, 15, 24, 24, 5242881, None,
                    "py.ico", None, None)
    c = costing.pushbutton("Return", 102, 57, 56, 17, 3, "Return", None)
    c.event("EndDialog", "Exit")

    #####################################################################
    # Preparation dialog: no user input except cancellation
    prep = PyDialog(db, "PrepareDlg", x, y, w, h, modeless, title,
                    "Cancel", "Cancel", "Cancel")
    prep.text("Description", 135, 70, 220, 40, 0x30003,
              "Please wait while the Installer prepares to guide you through the installation.")
    prep.title("Welcome to the [ProductName] Installer")
    c=prep.text("ActionText", 135, 110, 220, 20, 0x30003, "Pondering...")
    c.mapping("ActionText", "Text")
    c=prep.text("ActionData", 135, 135, 220, 30, 0x30003, None)
    c.mapping("ActionData", "Text")
    prep.back("Back", None, active=0)
    prep.next("Next", None, active=0)
    c=prep.cancel("Cancel", None)
    c.event("SpawnDialog", "CancelDlg")

    #####################################################################
    # Target directory selection
    seldlg = PyDialog(db, "SelectDirectoryDlg", x, y, w, h, modal, title,
                    "Next", "Next", "Cancel")
    seldlg.title("Select Destination Directory")
    seldlg.text("Description", 135, 50, 220, 40, 0x30003,
               "Please select a directory for the [ProductName] files.")

    seldlg.back("< Back", None, active=0)
    c = seldlg.next("Next >", "Cancel")
    c.event("DoAction", "CheckDir", "TargetExistsOk<>1", order=1)
    # If the target exists, but we found that we are going to remove old versions, don't bother
    # confirming that the target directory exists. Strictly speaking, we should determine that
    # the target directory is indeed the target of the product that we are going to remove, but
    # I don't know how to do that.
    c.event("SpawnDialog", "ExistingDirectoryDlg", 'TargetExists=1 and REMOVEOLDVERSION="" and REMOVEOLDSNAPSHOT=""', 2)
    c.event("SetTargetPath", "TARGETDIR", 'TargetExists=0 or REMOVEOLDVERSION<>"" or REMOVEOLDSNAPSHOT<>""', 3)
    c.event("SpawnWaitDialog", "WaitForCostingDlg", "CostingComplete=1", 4)
    c.event("NewDialog", "SelectFeaturesDlg", 'TargetExists=0 or REMOVEOLDVERSION<>"" or REMOVEOLDSNAPSHOT<>""', 5)

    c = seldlg.cancel("Cancel", "DirectoryCombo")
    c.event("SpawnDialog", "CancelDlg")

    seldlg.control("DirectoryCombo", "DirectoryCombo", 135, 70, 172, 80, 393219,
                   "TARGETDIR", None, "DirectoryList", None)
    seldlg.control("DirectoryList", "DirectoryList", 135, 90, 208, 136, 3, "TARGETDIR",
                   None, "PathEdit", None)
    seldlg.control("PathEdit", "PathEdit", 135, 230, 206, 16, 3, "TARGETDIR", None, "Next", None)
    c = seldlg.pushbutton("Up", 306, 70, 18, 18, 3, "Up", None)
    c.event("DirectoryListUp", "0")
    c = seldlg.pushbutton("NewDir", 324, 70, 30, 18, 3, "New", None)
    c.event("DirectoryListNew", "0")

    #####################################################################
    # SelectFeaturesDlg
    features = PyDialog(db, "SelectFeaturesDlg", x, y, w, h, modal|track_disk_space,
                        title, "Tree", "Next", "Cancel")
    features.title("Customize [ProductName]")
    features.text("Description", 135, 35, 220, 15, 0x30003,
                  "Select the way you want features to be installed.")
    features.text("Text", 135,45,220,30, 3,
                  "Click on the icons in the tree below to change the way features will be installed.")

    c=features.back("< Back", "Next")
    c.event("NewDialog", "SelectDirectoryDlg")

    c=features.next("Next >", "Cancel")
    c.mapping("SelectionNoItems", "Enabled")
    c.event("SpawnDialog", "DiskCostDlg", "OutOfDiskSpace=1", order=1)
    c.event("EndDialog", "Return", "OutOfDiskSpace<>1", order=2)

    c=features.cancel("Cancel", "Tree")
    c.event("SpawnDialog", "CancelDlg")

    # The browse property is not used, since we have only a single target path (selected already)
    features.control("Tree", "SelectionTree", 135, 75, 220, 95, 7, "_BrowseProperty",
                     "Tree of selections", "Back", None)

    #c=features.pushbutton("Reset", 42, 243, 56, 17, 3, "Reset", "DiskCost")
    #c.mapping("SelectionNoItems", "Enabled")
    #c.event("Reset", "0")

    features.control("Box", "GroupBox", 135, 170, 225, 90, 1, None, None, None, None)

    c=features.xbutton("DiskCost", "Disk &Usage", None, 0.10)
    c.mapping("SelectionNoItems","Enabled")
    c.event("SpawnDialog", "DiskCostDlg")

    c=features.xbutton("Advanced", "Advanced", None, 0.30)
    c.event("SpawnDialog", "AdvancedDlg")

    c=features.text("ItemDescription", 140, 180, 210, 30, 3,
                  "Multiline description of the currently selected item.")
    c.mapping("SelectionDescription","Text")

    c=features.text("ItemSize", 140, 210, 210, 45, 3,
                    "The size of the currently selected item.")
    c.mapping("SelectionSize", "Text")

    #####################################################################
    # Disk cost
    cost = PyDialog(db, "DiskCostDlg", x, y, w, h, modal, title,
                    "OK", "OK", "OK", bitmap=False)
    cost.text("Title", 15, 6, 200, 15, 0x30003,
              "{\DlgFontBold8}Disk Space Requirements")
    cost.text("Description", 20, 20, 280, 20, 0x30003,
              "The disk space required for the installation of the selected features.")
    cost.text("Text", 20, 53, 330, 60, 3,
              "The highlighted volumes (if any) do not have enough disk space "
              "available for the currently selected features.  You can either "
              "remove some files from the highlighted volumes, or choose to "
              "install less features onto local drive(s), or select different "
              "destination drive(s).")
    cost.control("VolumeList", "VolumeCostList", 20, 100, 330, 150, 393223,
                 None, "{120}{70}{70}{70}{70}", None, None)
    cost.xbutton("OK", "Ok", None, 0.5).event("EndDialog", "Return")

    #####################################################################
    # WhichUsers Dialog. Only available on NT, and for privileged users.
    # This must be run before FindRelatedProducts, because that will
    # take into account whether the previous installation was per-user
    # or per-machine. We currently don't support going back to this
    # dialog after "Next" was selected; to support this, we would need to
    # find how to reset the ALLUSERS property, and how to re-run
    # FindRelatedProducts.
    # On Windows9x, the ALLUSERS property is ignored on the command line
    # and in the Property table, but installer fails according to the documentation
    # if a dialog attempts to set ALLUSERS.
    whichusers = PyDialog(db, "WhichUsersDlg", x, y, w, h, modal, title,
                        "AdminInstall", "Next", "Cancel")
    whichusers.title("Select whether to install [ProductName] for all users of this computer.")
    # A radio group with two options: allusers, justme
    g = whichusers.radiogroup("AdminInstall", 135, 60, 160, 50, 3,
                              "WhichUsers", "", "Next")
    g.add("ALL", 0, 5, 150, 20, "Install for all users")
    g.add("JUSTME", 0, 25, 150, 20, "Install just for me")

    whichusers.back("Back", None, active=0)

    c = whichusers.next("Next >", "Cancel")
    c.event("[ALLUSERS]", "1", 'WhichUsers="ALL"', 1)
    c.event("EndDialog", "Return", order = 2)

    c = whichusers.cancel("Cancel", "AdminInstall")
    c.event("SpawnDialog", "CancelDlg")

    #####################################################################
    # Advanced Dialog.
    advanced = PyDialog(db, "AdvancedDlg", x, y, w, h, modal, title,
                        "CompilePyc", "Next", "Cancel")
    advanced.title("Advanced Options for [ProductName]")
    # A radio group with two options: allusers, justme
    advanced.checkbox("CompilePyc", 135, 60, 230, 50, 3,
                      "COMPILEALL", "Compile .py files to byte code after installation", "Next")

    c = advanced.next("Finish", "Cancel")
    c.event("EndDialog", "Return")

    c = advanced.cancel("Cancel", "CompilePyc")
    c.event("SpawnDialog", "CancelDlg")

    #####################################################################
    # Existing Directory dialog
    dlg = Dialog(db, "ExistingDirectoryDlg", 50, 30, 200, 80, modal, title,
                   "No", "No", "No")
    dlg.text("Title", 10, 20, 180, 40, 3,
             "[TARGETDIR] exists. Are you sure you want to overwrite existing files?")
    c=dlg.pushbutton("Yes", 30, 60, 55, 17, 3, "Yes", "No")
    c.event("[TargetExists]", "0", order=1)
    c.event("[TargetExistsOk]", "1", order=2)
    c.event("EndDialog", "Return", order=3)
    c=dlg.pushbutton("No", 115, 60, 55, 17, 3, "No", "Yes")
    c.event("EndDialog", "Return")

    #####################################################################
    # Installation Progress dialog (modeless)
    progress = PyDialog(db, "ProgressDlg", x, y, w, h, modeless, title,
                        "Cancel", "Cancel", "Cancel", bitmap=False)
    progress.text("Title", 20, 15, 200, 15, 0x30003,
                  "{\DlgFontBold8}[Progress1] [ProductName]")
    progress.text("Text", 35, 65, 300, 30, 3,
                  "Please wait while the Installer [Progress2] [ProductName]. "
                  "This may take several minutes.")
    progress.text("StatusLabel", 35, 100, 35, 20, 3, "Status:")

    c=progress.text("ActionText", 70, 100, w-70, 20, 3, "Pondering...")
    c.mapping("ActionText", "Text")

    #c=progress.text("ActionData", 35, 140, 300, 20, 3, None)
    #c.mapping("ActionData", "Text")

    c=progress.control("ProgressBar", "ProgressBar", 35, 120, 300, 10, 65537,
                       None, "Progress done", None, None)
    c.mapping("SetProgress", "Progress")

    progress.back("< Back", "Next", active=False)
    progress.next("Next >", "Cancel", active=False)
    progress.cancel("Cancel", "Back").event("SpawnDialog", "CancelDlg")

    # Maintenance type: repair/uninstall
    maint = PyDialog(db, "MaintenanceTypeDlg", x, y, w, h, modal, title,
                     "Next", "Next", "Cancel")
    maint.title("Welcome to the [ProductName] Setup Wizard")
    maint.text("BodyText", 135, 63, 230, 42, 3,
               "Select whether you want to repair or remove [ProductName].")
    g=maint.radiogroup("RepairRadioGroup", 135, 108, 230, 60, 3,
                        "MaintenanceForm_Action", "", "Next")
    g.add("Change", 0, 0, 200, 17, "&Change [ProductName]")
    g.add("Repair", 0, 18, 200, 17, "&Repair [ProductName]")
    g.add("Remove", 0, 36, 200, 17, "Re&move [ProductName]")

    maint.back("< Back", None, active=False)
    c=maint.next("Finish", "Cancel")
    # Change installation: Change progress dialog to "Change", then ask
    # for feature selection
    c.event("[Progress1]", "Change", 'MaintenanceForm_Action="Change"', 1)
    c.event("[Progress2]", "changes", 'MaintenanceForm_Action="Change"', 2)

    # Reinstall: Change progress dialog to "Repair", then invoke reinstall
    # Also set list of reinstalled features to "ALL"
    c.event("[REINSTALL]", "ALL", 'MaintenanceForm_Action="Repair"', 5)
    c.event("[Progress1]", "Repairing", 'MaintenanceForm_Action="Repair"', 6)
    c.event("[Progress2]", "repaires", 'MaintenanceForm_Action="Repair"', 7)
    c.event("Reinstall", "ALL", 'MaintenanceForm_Action="Repair"', 8)

    # Uninstall: Change progress to "Remove", then invoke uninstall
    # Also set list of removed features to "ALL"
    c.event("[REMOVE]", "ALL", 'MaintenanceForm_Action="Remove"', 11)
    c.event("[Progress1]", "Removing", 'MaintenanceForm_Action="Remove"', 12)
    c.event("[Progress2]", "removes", 'MaintenanceForm_Action="Remove"', 13)
    c.event("Remove", "ALL", 'MaintenanceForm_Action="Remove"', 14)

    # Close dialog when maintenance action scheduled
    c.event("EndDialog", "Return", 'MaintenanceForm_Action<>"Change"', 20)
    c.event("NewDialog", "SelectFeaturesDlg", 'MaintenanceForm_Action="Change"', 21)

    maint.cancel("Cancel", "RepairRadioGroup").event("SpawnDialog", "CancelDlg")


# See "Feature Table". The feature level is 1 for all features,
# and the feature attributes are 0 for the DefaultFeature, and
# FollowParent for all other features. The numbers are the Display
# column.
def add_features(db):
    # feature attributes:
    # msidbFeatureAttributesFollowParent == 2
    # msidbFeatureAttributesDisallowAdvertise == 8
    # Features that need to be installed with together with the main feature
    # (i.e. additional Python libraries) need to follow the parent feature.
    # Features that have no advertisement trigger (e.g. the test suite)
    # must not support advertisement
    global default_feature, tcltk, htmlfiles, tools, testsuite, ext_feature
    default_feature = Feature(db, "DefaultFeature", "Python",
                              "Python Interpreter and Libraries",
                              1, directory = "TARGETDIR")
    # The extensions feature is tricky wrt. advertisement and follow parent.
    # The following combinations must be supported:
    # default feature         extensions                       effect on extensions
    # locally/from source     locally/from source              registered
    # advertised              advertised/locally/from source   advertised
    # locally/from source     not installed                    not installed
    # advertised              not installed                    not advertised
    #                                                          (only shortcuts are)
    # The following combination might be considered meaningless, but cannot be excluded
    # locally/from source     advertised                       registered
    ext_feature = Feature(db, "Extensions", "Register Extensions",
                          "Make this Python installation the default Python installation", 3,
                         parent = default_feature)
    if have_tcl:
        tcltk = Feature(db, "TclTk", "Tcl/Tk", "Tkinter, IDLE, pydoc", 5,
                    parent = default_feature, attributes=2)
    htmlfiles = Feature(db, "Documentation", "Documentation",
                        "Python HTMLHelp File", 7, parent = default_feature)
    tools = Feature(db, "Tools", "Utility Scripts",
                    "Python utility scripts (Tools/", 9,
                    parent = default_feature, attributes=2)
    testsuite = Feature(db, "Testsuite", "Test suite",
                        "Python test suite (Lib/test/)", 11,
                        parent = default_feature, attributes=2|8)


def extract_msvcr71():
    import _winreg
    # Find the location of the merge modules
    k = _winreg.OpenKey(
        _winreg.HKEY_LOCAL_MACHINE,
        r"Software\Microsoft\VisualStudio\7.1\Setup\VS")
    dir = _winreg.QueryValueEx(k, "MSMDir")[0]
    _winreg.CloseKey(k)
    files = glob.glob1(dir, "*CRT71*")
    assert len(files) == 1
    file = os.path.join(dir, files[0])
    # Extract msvcr71.dll
    m = msilib.MakeMerge2()
    m.OpenModule(file, 0)
    m.ExtractFiles(".")
    m.CloseModule()
    # Find the version/language of msvcr71.dll
    installer = msilib.MakeInstaller()
    return installer.FileVersion("msvcr71.dll", 0), \
           installer.FileVersion("msvcr71.dll", 1)

class PyDirectory(Directory):
    """By default, all components in the Python installer
    can run from source."""
    def __init__(self, *args, **kw):
        if not kw.has_key("componentflags"):
            kw['componentflags'] = 2 #msidbComponentAttributesOptional
        Directory.__init__(self, *args, **kw)

# See "File Table", "Component Table", "Directory Table",
# "FeatureComponents Table"
def add_files(db):
    cab = CAB("python")
    tmpfiles = []
    # Add all executables, icons, text files into the TARGETDIR component
    root = PyDirectory(db, cab, None, srcdir, "TARGETDIR", "SourceDir")
    default_feature.set_current()
    if not msilib.Win64:
        root.add_file("PCBuild/w9xpopen.exe")
    root.add_file("PC/py.ico")
    root.add_file("PC/pyc.ico")
    root.add_file("README.txt", src="README")
    root.add_file("NEWS.txt", src="Misc/NEWS")
    root.add_file("LICENSE.txt", src="LICENSE")
    root.start_component("python.exe", keyfile="python.exe")
    root.add_file("PCBuild/python.exe")
    root.start_component("pythonw.exe", keyfile="pythonw.exe")
    root.add_file("PCBuild/pythonw.exe")

    # msidbComponentAttributesSharedDllRefCount = 8, see "Component Table"
    dlldir = PyDirectory(db, cab, root, srcdir, "DLLDIR", ".")
    pydll = "python%s%s.dll" % (major, minor)
    pydllsrc = srcdir + "/PCBuild/" + pydll
    dlldir.start_component("DLLDIR", flags = 8, keyfile = pydll)
    installer = msilib.MakeInstaller()
    pyversion = installer.FileVersion(pydllsrc, 0)
    if not snapshot:
        # For releases, the Python DLL has the same version as the
        # installer package.
        assert pyversion.split(".")[:3] == current_version.split(".")
    dlldir.add_file("PCBuild/python%s%s.dll" % (major, minor),
                    version=pyversion,
                    language=installer.FileVersion(pydllsrc, 1))
    # XXX determine dependencies
    version, lang = extract_msvcr71()
    dlldir.start_component("msvcr71", flags=8, keyfile="msvcr71.dll")
    dlldir.add_file("msvcr71.dll", src=os.path.abspath("msvcr71.dll"),
                    version=version, language=lang)
    tmpfiles.append("msvcr71.dll")

    # Add all .py files in Lib, except lib-tk, test
    dirs={}
    pydirs = [(root,"Lib")]
    while pydirs:
        parent, dir = pydirs.pop()
        if dir == "CVS" or dir.startswith("plat-"):
            continue
        elif dir in ["lib-tk", "idlelib", "Icons"]:
            if not have_tcl:
                continue
            tcltk.set_current()
        elif dir in ['test', 'output']:
            testsuite.set_current()
        else:
            default_feature.set_current()
        lib = PyDirectory(db, cab, parent, dir, dir, "%s|%s" % (parent.make_short(dir), dir))
        # Add additional files
        dirs[dir]=lib
        lib.glob("*.txt")
        if dir=='site-packages':
            continue
        files = lib.glob("*.py")
        files += lib.glob("*.pyw")
        if files:
            # Add an entry to the RemoveFile table to remove bytecode files.
            lib.remove_pyc()
        if dir=='test' and parent.physical=='Lib':
            lib.add_file("185test.db")
            lib.add_file("audiotest.au")
            lib.add_file("cfgparser.1")
            lib.add_file("test.xml")
            lib.add_file("test.xml.out")
            lib.add_file("testtar.tar")
            lib.glob("*.uue")
            lib.add_file("readme.txt", src="README")
        if dir=='decimaltestdata':
            lib.glob("*.decTest")
        if dir=='output':
            lib.glob("test_*")
        if dir=='idlelib':
            lib.glob("*.def")
            lib.add_file("idle.bat")
        if dir=="Icons":
            lib.glob("*.gif")
            lib.add_file("idle.icns")
        if dir=="command":
            lib.add_file("wininst-6.exe")
            lib.add_file("wininst-7.1.exe")
        if dir=="data" and parent.physical=="test" and parent.basedir.physical=="email":
            # This should contain all non-CVS files listed in CVS
            for f in os.listdir(lib.absolute):
                if f.endswith(".txt") or f=="CVS":continue
                if f.endswith(".au") or f.endswith(".gif"):
                    lib.add_file(f)
                else:
                    print "WARNING: New file %s in email/test/data" % f
        for f in os.listdir(lib.absolute):
            if os.path.isdir(os.path.join(lib.absolute, f)):
                pydirs.append((lib, f))
    # Add DLLs
    default_feature.set_current()
    lib = PyDirectory(db, cab, root, srcdir+"/PCBuild", "DLLs", "DLLS|DLLs")
    dlls = []
    tclfiles = []
    for f in extensions:
        if f=="_tkinter.pyd":
            continue
        if not os.path.exists(srcdir+"/PCBuild/"+f):
            print "WARNING: Missing extension", f
            continue
        dlls.append(f)
        lib.add_file(f)
    if have_tcl:
        if not os.path.exists(srcdir+"/PCBuild/_tkinter.pyd"):
            print "WARNING: Missing _tkinter.pyd"
        else:
            lib.start_component("TkDLLs", tcltk)
            lib.add_file("_tkinter.pyd")
            dlls.append("_tkinter.pyd")
            tcldir = os.path.normpath(srcdir+"/../tcltk/bin")
            for f in glob.glob1(tcldir, "*.dll"):
                lib.add_file(f, src=os.path.join(tcldir, f))
    # check whether there are any unknown extensions
    for f in glob.glob1(srcdir+"/PCBuild", "*.pyd"):
        if f.endswith("_d.pyd"): continue # debug version
        if f in dlls: continue
        print "WARNING: Unknown extension", f

    # Add headers
    default_feature.set_current()
    lib = PyDirectory(db, cab, root, "include", "include", "INCLUDE|include")
    lib.glob("*.h")
    lib.add_file("pyconfig.h", src="../PC/pyconfig.h")
    # Add import libraries
    lib = PyDirectory(db, cab, root, "PCBuild", "libs", "LIBS|libs")
    for f in dlls:
        lib.add_file(f.replace('pyd','lib'))
    lib.add_file('python%s%s.lib' % (major, minor))
    if have_tcl:
        # Add Tcl/Tk
        tcldirs = [(root, '../tcltk/lib', 'tcl')]
        tcltk.set_current()
        while tcldirs:
            parent, phys, dir = tcldirs.pop()
            lib = PyDirectory(db, cab, parent, phys, dir, "%s|%s" % (parent.make_short(dir), dir))
            if not os.path.exists(lib.absolute):
                continue
            for f in os.listdir(lib.absolute):
                if os.path.isdir(os.path.join(lib.absolute, f)):
                    tcldirs.append((lib, f, f))
                else:
                    lib.add_file(f)
    # Add tools
    tools.set_current()
    tooldir = PyDirectory(db, cab, root, "Tools", "Tools", "TOOLS|Tools")
    for f in ['i18n', 'pynche', 'Scripts', 'versioncheck', 'webchecker']:
        lib = PyDirectory(db, cab, tooldir, f, f, "%s|%s" % (tooldir.make_short(f), f))
        lib.glob("*.py")
        lib.glob("*.pyw", exclude=['pydocgui.pyw'])
        lib.remove_pyc()
        lib.glob("*.txt")
        if f == "pynche":
            x = PyDirectory(db, cab, lib, "X", "X", "X|X")
            x.glob("*.txt")
        if f == 'Scripts':
            lib.add_file("README.txt", src="README")
            if have_tcl:
                lib.start_component("pydocgui.pyw", tcltk, keyfile="pydocgui.pyw")
                lib.add_file("pydocgui.pyw")
    # Add documentation
    htmlfiles.set_current()
    lib = PyDirectory(db, cab, root, "Doc", "Doc", "DOC|Doc")
    lib.start_component("documentation", keyfile="Python%s%s.chm" % (major,minor))
    lib.add_file("Python%s%s.chm" % (major, minor))

    cab.commit(db)

    for f in tmpfiles:
        os.unlink(f)

# See "Registry Table", "Component Table"
def add_registry(db):
    # File extensions, associated with the REGISTRY.def component
    # IDLE verbs depend on the tcltk feature.
    # msidbComponentAttributesRegistryKeyPath = 4
    # -1 for Root specifies "dependent on ALLUSERS property"
    tcldata = []
    if have_tcl:
        tcldata = [
            ("REGISTRY.tcl", msilib.gen_uuid(), "TARGETDIR", 4,
               "&%s <> 2" % ext_feature.id, "py.IDLE")]
    add_data(db, "Component",
             # msidbComponentAttributesRegistryKeyPath = 4
             [("REGISTRY", msilib.gen_uuid(), "TARGETDIR", 4, None,
               "InstallPath"),
              ("REGISTRY.def", msilib.gen_uuid(), "TARGETDIR", 4,
               None, None)] + tcldata)
    # See "FeatureComponents Table".
    # The association between TclTk and pythonw.exe is necessary to make ICE59
    # happy, because the installer otherwise believes that the IDLE and PyDoc
    # shortcuts might get installed without pythonw.exe being install. This
    # is not true, since installing TclTk will install the default feature, which
    # will cause pythonw.exe to be installed.
    tcldata = []
    if have_tcl:
        tcltkdata = [(tcltk.id, "REGISTRY.tcl"),
                     (tcltk.id, "pythonw.exe")]
    add_data(db, "FeatureComponents",
             [(default_feature.id, "REGISTRY"),
              (ext_feature.id, "REGISTRY.def")] +
              tcldata
              )

    pat = r"Software\Classes\%sPython.%sFile\shell\%s\command"
    ewi = "Edit with IDLE"
    pat2 = r"Software\Classes\%sPython.%sFile\DefaultIcon"
    pat3 = r"Software\Classes\%sPython.%sFile"
    # Advertised extensions
    add_data(db, "Extension",
            [("py", "python.exe", "Python.File", None, default_feature.id),
            ("pyw", "pythonw.exe", "Python.NoConFile", None, default_feature.id),
            ("pyc", "python.exe", "Python.CompiledFile", None, default_feature.id),
            ("pyo", "python.exe", "Python.CompiledFile", None, default_feature.id)])
    # add_data(db, "MIME") XXX
    add_data(db, "Verb",
            [("py", "open", 1, None, r'-n -e "%1"'),
            ("pyw", "open", 1, None, r'-n -e "%1"'),
            ("pyc", "open", 1, None, r'-n -e "%1"'),
            ("pyo", "open", 1, None, r'-n -e "%1"')])
    add_data(db, "ProgId",
            [("Python.File", None, None, "Python File", "python_icon.exe", 0),
             ("Python.NoConFile", None, None, "Python File (no console)", "python_icon.exe", 0),
             ("Python.CompiledFile", None, None, "Compiled Python File", "python_icon.exe", 1)])

    # Non-advertised verbs: for advertised verbs, we would need to invoke the same
    # executable for both open and "Edit with IDLE". This cannot work, as we want
    # to use pythonw.exe in either case
    if have_tcl:
        add_data(db, "Registry",
            [#Verbs
             ("py.IDLE", -1, pat % (testprefix, "", ewi), "",
              r'"[TARGETDIR]pythonw.exe" "[TARGETDIR]Lib\idlelib\idle.pyw" -n -e "%1"',
              "REGISTRY.tcl"),
             ("pyw.IDLE", -1, pat % (testprefix, "NoCon", ewi), "",
              r'"[TARGETDIR]pythonw.exe" "[TARGETDIR]Lib\idlelib\idle.pyw" -n -e "%1"',
              "REGISTRY.tcl"),
            ])

    # Registry keys
    prefix = r"Software\%sPython\PythonCore\%s" % (testprefix, short_version)
    add_data(db, "Registry",
             [("InstallPath", -1, prefix+r"\InstallPath", "", "[TARGETDIR]", "REGISTRY"),
              ("InstallGroup", -1, prefix+r"\InstallPath\InstallGroup", "",
               "Python %s" % short_version, "REGISTRY"),
              ("PythonPath", -1, prefix+r"\PythonPath", "",
               "[TARGETDIR]Lib;[TARGETDIR]DLLs;[TARGETDIR]lib-tk", "REGISTRY"),
              ("Documentation", -1, prefix+r"\Help\Main Python Documentation", "",
               r"[TARGETDIR]Doc\Python%s%s.chm" % (major, minor), "REGISTRY"),
              ("Modules", -1, prefix+r"\Modules", "+", None, "REGISTRY"),
              ("AppPaths", -1, r"Software\Microsoft\Windows\CurrentVersion\App Paths\Python.exe",
               "", r"[TARGETDIR]Python.exe", "REGISTRY.def")
              ])
    # Shortcuts, see "Shortcut Table"
    add_data(db, "Directory",
             [("ProgramMenuFolder", "TARGETDIR", "."),
              ("MenuDir", "ProgramMenuFolder", "PY%s%s|%sPython %s.%s" % (major,minor,testprefix,major,minor))])
    add_data(db, "RemoveFile",
             [("MenuDir", "TARGETDIR", None, "MenuDir", 2)])
    tcltkshortcuts = []
    if have_tcl:
        tcltkshortcuts = [
              ("IDLE", "MenuDir", "IDLE|IDLE (Python GUI)", "pythonw.exe",
               tcltk.id, r"[TARGETDIR]Lib\idlelib\idle.pyw", None, None, "python_icon.exe", 0, None, "TARGETDIR"),
              ("PyDoc", "MenuDir", "MODDOCS|Module Docs", "pythonw.exe",
               tcltk.id, r"[TARGETDIR]Tools\scripts\pydocgui.pyw", None, None, "python_icon.exe", 0, None, "TARGETDIR"),
              ]
    add_data(db, "Shortcut",
             tcltkshortcuts +
             [# Advertised shortcuts: targets are features, not files
              ("Python", "MenuDir", "PYTHON|Python (command line)", "python.exe",
               default_feature.id, None, None, None, "python_icon.exe", 2, None, "TARGETDIR"),
              ("Manual", "MenuDir", "MANUAL|Python Manuals", "documentation",
               htmlfiles.id, None, None, None, None, None, None, None),
              ## Non-advertised shortcuts: must be associated with a registry component
              ("Uninstall", "MenuDir", "UNINST|Uninstall Python", "REGISTRY",
               SystemFolderName+"msiexec",  "/x%s" % product_code,
               None, None, None, None, None, None),
              ])
    db.Commit()

db = build_database()
try:
    add_features(db)
    add_ui(db)
    add_files(db)
    add_registry(db)
    remove_old_versions(db)
    db.Commit()
finally:
    del db



# A minimal text editor.
#
# To be done:
# - Functionality: find, etc.

from Carbon.Menu import DrawMenuBar
from FrameWork import *
from Carbon import Win
from Carbon import Qd
from Carbon import Res
from Carbon import Fm
import waste
import WASTEconst
from Carbon import Scrap
import os
import EasyDialogs
import macfs
import string
import htmllib

WATCH = Qd.GetCursor(4).data

LEFTMARGIN=0

UNDOLABELS = [ # Indexed by WEGetUndoInfo() value
        None, "", "typing", "Cut", "Paste", "Clear", "Drag", "Style"]

# Style and size menu. Note that style order is important (tied to bit values)
STYLES = [
        ("Bold", "B"), ("Italic", "I"), ("Underline", "U"), ("Outline", "O"),
        ("Shadow", ""), ("Condensed", ""), ("Extended", "")
        ]
SIZES = [ 9, 10, 12, 14, 18, 24]

# Sizes for HTML tag types
HTML_SIZE={
        'h1': 18,
        'h2': 14
}

BIGREGION=Qd.NewRgn()
Qd.SetRectRgn(BIGREGION, -16000, -16000, 16000, 16000)

class WasteWindow(ScrolledWindow):
    def open(self, path, name, data):
        self.path = path
        self.name = name
        r = windowbounds(400, 400)
        w = Win.NewWindow(r, name, 1, 0, -1, 1, 0)
        self.wid = w
        vr = LEFTMARGIN, 0, r[2]-r[0]-15, r[3]-r[1]-15
        dr = (0, 0, vr[2], 0)
        Qd.SetPort(w)
        Qd.TextFont(4)
        Qd.TextSize(9)
        flags = WASTEconst.weDoAutoScroll | WASTEconst.weDoOutlineHilite | \
                WASTEconst.weDoMonoStyled | WASTEconst.weDoUndo
        self.ted = waste.WENew(dr, vr, flags)
        self.ted.WEInstallTabHooks()
        style, soup = self.getstylesoup(self.path)
        self.ted.WEInsert(data, style, soup)
        self.ted.WESetSelection(0,0)
        self.ted.WECalText()
        self.ted.WEResetModCount()
        w.DrawGrowIcon()
        self.scrollbars()
        self.do_postopen()
        self.do_activate(1, None)

    def getstylesoup(self, pathname):
        if not pathname:
            return None, None
        oldrf = Res.CurResFile()
        try:
            rf = Res.FSpOpenResFile(self.path, 1)
        except Res.Error:
            return None, None
        try:
            hstyle = Res.Get1Resource('styl', 128)
            hstyle.DetachResource()
        except Res.Error:
            hstyle = None
        try:
            hsoup = Res.Get1Resource('SOUP', 128)
            hsoup.DetachResource()
        except Res.Error:
            hsoup = None
        Res.CloseResFile(rf)
        Res.UseResFile(oldrf)
        return hstyle, hsoup

    def do_idle(self, event):
        (what, message, when, where, modifiers) = event
        Qd.SetPort(self.wid)
        self.ted.WEIdle()
        if self.ted.WEAdjustCursor(where, BIGREGION):
            return
        Qd.SetCursor(Qd.GetQDGlobalsArrow())

    def getscrollbarvalues(self):
        dr = self.ted.WEGetDestRect()
        vr = self.ted.WEGetViewRect()
        vx = self.scalebarvalue(dr[0], dr[2], vr[0], vr[2])
        vy = self.scalebarvalue(dr[1], dr[3], vr[1], vr[3])
        return vx, vy

    def scrollbar_callback(self, which, what, value):
        if which == 'y':
            #
            # "line" size is minimum of top and bottom line size
            #
            topline_off,dummy = self.ted.WEGetOffset((1,1))
            topline_num = self.ted.WEOffsetToLine(topline_off)
            toplineheight = self.ted.WEGetHeight(topline_num, topline_num+1)

            botlinepos = self.ted.WEGetViewRect()[3]
            botline_off, dummy = self.ted.WEGetOffset((1, botlinepos-1))
            botline_num = self.ted.WEOffsetToLine(botline_off)
            botlineheight = self.ted.WEGetHeight(botline_num, botline_num+1)

            if botlineheight == 0:
                botlineheight = self.ted.WEGetHeight(botline_num-1, botline_num)
            if botlineheight < toplineheight:
                lineheight = botlineheight
            else:
                lineheight = toplineheight
            if lineheight <= 0:
                lineheight = 1
            #
            # Now do the command.
            #
            if what == 'set':
                height = self.ted.WEGetHeight(0, 0x3fffffff)
                cur = self.getscrollbarvalues()[1]
                delta = (cur-value)*height/32767
            if what == '-':
                delta = lineheight
            elif what == '--':
                delta = (self.ted.WEGetViewRect()[3]-lineheight)
                if delta <= 0:
                    delta = lineheight
            elif what == '+':
                delta = -lineheight
            elif what == '++':
                delta = -(self.ted.WEGetViewRect()[3]-lineheight)
                if delta >= 0:
                    delta = -lineheight
            self.ted.WEScroll(0, delta)
        else:
            if what == 'set':
                return # XXXX
            vr = self.ted.WEGetViewRect()
            winwidth = vr[2]-vr[0]
            if what == '-':
                delta = winwidth/10
            elif what == '--':
                delta = winwidth/2
            elif what == '+':
                delta = -winwidth/10
            elif what == '++':
                delta = -winwidth/2
            self.ted.WEScroll(delta, 0)
        # Pin the scroll
        l, t, r, b = self.ted.WEGetDestRect()
        vl, vt, vr, vb = self.ted.WEGetViewRect()
        if t > 0 or l > 0:
            dx = dy = 0
            if t > 0: dy = -t
            if l > 0: dx = -l
            self.ted.WEScroll(dx, dy)
        elif b < vb:
            self.ted.WEScroll(0, vb-b)


    def do_activate(self, onoff, evt):
        Qd.SetPort(self.wid)
        ScrolledWindow.do_activate(self, onoff, evt)
        if onoff:
            self.ted.WEActivate()
            self.parent.active = self
            self.parent.updatemenubar()
        else:
            self.ted.WEDeactivate()

    def do_update(self, wid, event):
        region = wid.GetWindowPort().visRgn
        if Qd.EmptyRgn(region):
            return
        Qd.EraseRgn(region)
        self.ted.WEUpdate(region)
        self.updatescrollbars()

    def do_postresize(self, width, height, window):
        l, t, r, b = self.ted.WEGetViewRect()
        vr = (l, t, l+width-15, t+height-15)
        self.ted.WESetViewRect(vr)
        self.wid.InvalWindowRect(vr)
        ScrolledWindow.do_postresize(self, width, height, window)

    def do_contentclick(self, local, modifiers, evt):
        (what, message, when, where, modifiers) = evt
        self.ted.WEClick(local, modifiers, when)
        self.updatescrollbars()
        self.parent.updatemenubar()

    def do_char(self, ch, event):
        self.ted.WESelView()
        (what, message, when, where, modifiers) = event
        self.ted.WEKey(ord(ch), modifiers)
        self.updatescrollbars()
        self.parent.updatemenubar()

    def close(self):
        if self.ted.WEGetModCount():
            save = EasyDialogs.AskYesNoCancel('Save window "%s" before closing?'%self.name, 1)
            if save > 0:
                self.menu_save()
            elif save < 0:
                return
        if self.parent.active == self:
            self.parent.active = None
        self.parent.updatemenubar()
        del self.ted
        self.do_postclose()

    def menu_save(self):
        if not self.path:
            self.menu_save_as()
            return # Will call us recursively
        #
        # First save data
        #
        dhandle = self.ted.WEGetText()
        data = dhandle.data
        fp = open(self.path, 'wb')  # NOTE: wb, because data has CR for end-of-line
        fp.write(data)
        if data[-1] <> '\r': fp.write('\r')
        fp.close()
        #
        # Now save style and soup
        #
        oldresfile = Res.CurResFile()
        try:
            rf = Res.FSpOpenResFile(self.path, 3)
        except Res.Error:
            Res.FSpCreateResFile(self.path, '????', 'TEXT', macfs.smAllScripts)
            rf = Res.FSpOpenResFile(self.path, 3)
        styles = Res.Resource('')
        soup = Res.Resource('')
        self.ted.WECopyRange(0, 0x3fffffff, None, styles, soup)
        styles.AddResource('styl', 128, '')
        soup.AddResource('SOUP', 128, '')
        Res.CloseResFile(rf)
        Res.UseResFile(oldresfile)

        self.ted.WEResetModCount()

    def menu_save_as(self):
        path = EasyDialogs.AskFileForSave(message='Save as:')
        if not path: return
        self.path = path
        self.name = os.path.split(self.path)[-1]
        self.wid.SetWTitle(self.name)
        self.menu_save()

    def menu_insert(self, fp):
        self.ted.WESelView()
        data = fp.read()
        self.ted.WEInsert(data, None, None)
        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_insert_html(self, fp):
        import htmllib
        import formatter
        f = formatter.AbstractFormatter(self)

        # Remember where we are, and don't update
        Qd.SetCursor(WATCH)
        start, dummy = self.ted.WEGetSelection()
        self.ted.WEFeatureFlag(WASTEconst.weFInhibitRecal, 1)

        self.html_init()
        p = MyHTMLParser(f)
        p.feed(fp.read())

        # Restore updating, recalc, set focus
        dummy, end = self.ted.WEGetSelection()
        self.ted.WECalText()
        self.ted.WESetSelection(start, end)
        self.ted.WESelView()
        self.ted.WEFeatureFlag(WASTEconst.weFInhibitRecal, 0)
        self.wid.InvalWindowRect(self.ted.WEGetViewRect())

        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_cut(self):
        self.ted.WESelView()
        if hasattr(Scrap, 'ZeroScrap'):
            Scrap.ZeroScrap()
        else:
            Scrap.ClearCurrentScrap()
        self.ted.WECut()
        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_copy(self):
        if hasattr(Scrap, 'ZeroScrap'):
            Scrap.ZeroScrap()
        else:
            Scrap.ClearCurrentScrap()
        self.ted.WECopy()
        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_paste(self):
        self.ted.WESelView()
        self.ted.WEPaste()
        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_clear(self):
        self.ted.WESelView()
        self.ted.WEDelete()
        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_undo(self):
        self.ted.WEUndo()
        self.updatescrollbars()
        self.parent.updatemenubar()

    def menu_setfont(self, font):
        font = Fm.GetFNum(font)
        self.mysetstyle(WASTEconst.weDoFont, (font, 0, 0, (0,0,0)))
        self.parent.updatemenubar()

    def menu_modface(self, face):
        self.mysetstyle(WASTEconst.weDoFace|WASTEconst.weDoToggleFace,
                (0, face, 0, (0,0,0)))

    def menu_setface(self, face):
        self.mysetstyle(WASTEconst.weDoFace|WASTEconst.weDoReplaceFace,
                (0, face, 0, (0,0,0)))

    def menu_setsize(self, size):
        self.mysetstyle(WASTEconst.weDoSize, (0, 0, size, (0,0,0)))

    def menu_incsize(self, size):
        self.mysetstyle(WASTEconst.weDoAddSize, (0, 0, size, (0,0,0)))

    def mysetstyle(self, which, how):
        self.ted.WESelView()
        self.ted.WESetStyle(which, how)
        self.parent.updatemenubar()

    def have_selection(self):
        start, stop = self.ted.WEGetSelection()
        return start < stop

    def can_paste(self):
        return self.ted.WECanPaste()

    def can_undo(self):
        which, redo = self.ted.WEGetUndoInfo()
        which = UNDOLABELS[which]
        if which == None: return None
        if redo:
            return "Redo "+which
        else:
            return "Undo "+which

    def getruninfo(self):
        all = (WASTEconst.weDoFont | WASTEconst.weDoFace | WASTEconst.weDoSize)
        dummy, mode, (font, face, size, color) = self.ted.WEContinuousStyle(all)
        if not (mode & WASTEconst.weDoFont):
            font = None
        else:
            font = Fm.GetFontName(font)
        if not (mode & WASTEconst.weDoFace): fact = None
        if not (mode & WASTEconst.weDoSize): size = None
        return font, face, size

    #
    # Methods for writer class for html formatter
    #

    def html_init(self):
        self.html_font = [12, 0, 0, 0]
        self.html_style = 0
        self.html_color = (0,0,0)
        self.new_font(self.html_font)

    def new_font(self, font):
        if font == None:
            font = (12, 0, 0, 0)
        font = map(lambda x:x, font)
        for i in range(len(font)):
            if font[i] == None:
                font[i] = self.html_font[i]
        [size, italic, bold, tt] = font
        self.html_font = font[:]
        if tt:
            font = Fm.GetFNum('Courier')
        else:
            font = Fm.GetFNum('Times')
        if HTML_SIZE.has_key(size):
            size = HTML_SIZE[size]
        else:
            size = 12
        face = 0
        if bold: face = face | 1
        if italic: face = face | 2
        face = face | self.html_style
        self.ted.WESetStyle(WASTEconst.weDoFont | WASTEconst.weDoFace |
                        WASTEconst.weDoSize | WASTEconst.weDoColor,
                        (font, face, size, self.html_color))

    def new_margin(self, margin, level):
        self.ted.WEInsert('[Margin %s %s]'%(margin, level), None, None)

    def new_spacing(self, spacing):
        self.ted.WEInsert('[spacing %s]'%spacing, None, None)

    def new_styles(self, styles):
        self.html_style = 0
        self.html_color = (0,0,0)
        if 'anchor' in styles:
            self.html_style = self.html_style | 4
            self.html_color = (0xffff, 0, 0)
        self.new_font(self.html_font)

    def send_paragraph(self, blankline):
        self.ted.WEInsert('\r'*(blankline+1), None, None)

    def send_line_break(self):
        self.ted.WEInsert('\r', None, None)

    def send_hor_rule(self, *args, **kw):
        # Ignore ruler options, for now
        dummydata = Res.Resource('')
        self.ted.WEInsertObject('rulr', dummydata, (0,0))

    def send_label_data(self, data):
        self.ted.WEInsert(data, None, None)

    def send_flowing_data(self, data):
        self.ted.WEInsert(data, None, None)

    def send_literal_data(self, data):
        data = string.replace(data, '\n', '\r')
        data = string.expandtabs(data)
        self.ted.WEInsert(data, None, None)

class Wed(Application):
    def __init__(self):
        Application.__init__(self)
        self.num = 0
        self.active = None
        self.updatemenubar()
        waste.STDObjectHandlers()
        # Handler for horizontal ruler
        waste.WEInstallObjectHandler('rulr', 'new ', self.newRuler)
        waste.WEInstallObjectHandler('rulr', 'draw', self.drawRuler)

    def makeusermenus(self):
        self.filemenu = m = Menu(self.menubar, "File")
        self.newitem = MenuItem(m, "New window", "N", self.open)
        self.openitem = MenuItem(m, "Open...", "O", self.openfile)
        self.closeitem = MenuItem(m, "Close", "W", self.closewin)
        m.addseparator()
        self.saveitem = MenuItem(m, "Save", "S", self.save)
        self.saveasitem = MenuItem(m, "Save as...", "", self.saveas)
        m.addseparator()
        self.insertitem = MenuItem(m, "Insert plaintext...", "", self.insertfile)
        self.htmlitem = MenuItem(m, "Insert HTML...", "", self.inserthtml)
        m.addseparator()
        self.quititem = MenuItem(m, "Quit", "Q", self.quit)

        self.editmenu = m = Menu(self.menubar, "Edit")
        self.undoitem = MenuItem(m, "Undo", "Z", self.undo)
        self.cutitem = MenuItem(m, "Cut", "X", self.cut)
        self.copyitem = MenuItem(m, "Copy", "C", self.copy)
        self.pasteitem = MenuItem(m, "Paste", "V", self.paste)
        self.clearitem = MenuItem(m, "Clear", "", self.clear)

        self.makefontmenu()

        # Groups of items enabled together:
        self.windowgroup = [self.closeitem, self.saveitem, self.saveasitem,
                self.editmenu, self.fontmenu, self.facemenu, self.sizemenu,
                self.insertitem]
        self.focusgroup = [self.cutitem, self.copyitem, self.clearitem]
        self.windowgroup_on = -1
        self.focusgroup_on = -1
        self.pastegroup_on = -1
        self.undo_label = "never"
        self.ffs_values = ()

    def makefontmenu(self):
        self.fontmenu = Menu(self.menubar, "Font")
        self.fontnames = getfontnames()
        self.fontitems = []
        for n in self.fontnames:
            m = MenuItem(self.fontmenu, n, "", self.selfont)
            self.fontitems.append(m)
        self.facemenu = Menu(self.menubar, "Style")
        self.faceitems = []
        for n, shortcut in STYLES:
            m = MenuItem(self.facemenu, n, shortcut, self.selface)
            self.faceitems.append(m)
        self.facemenu.addseparator()
        self.faceitem_normal = MenuItem(self.facemenu, "Normal", "N",
                self.selfacenormal)
        self.sizemenu = Menu(self.menubar, "Size")
        self.sizeitems = []
        for n in SIZES:
            m = MenuItem(self.sizemenu, repr(n), "", self.selsize)
            self.sizeitems.append(m)
        self.sizemenu.addseparator()
        self.sizeitem_bigger = MenuItem(self.sizemenu, "Bigger", "+",
                self.selsizebigger)
        self.sizeitem_smaller = MenuItem(self.sizemenu, "Smaller", "-",
                self.selsizesmaller)

    def selfont(self, id, item, *rest):
        if self.active:
            font = self.fontnames[item-1]
            self.active.menu_setfont(font)
        else:
            EasyDialogs.Message("No active window?")

    def selface(self, id, item, *rest):
        if self.active:
            face = (1<<(item-1))
            self.active.menu_modface(face)
        else:
            EasyDialogs.Message("No active window?")

    def selfacenormal(self, *rest):
        if self.active:
            self.active.menu_setface(0)
        else:
            EasyDialogs.Message("No active window?")

    def selsize(self, id, item, *rest):
        if self.active:
            size = SIZES[item-1]
            self.active.menu_setsize(size)
        else:
            EasyDialogs.Message("No active window?")

    def selsizebigger(self, *rest):
        if self.active:
            self.active.menu_incsize(2)
        else:
            EasyDialogs.Message("No active window?")

    def selsizesmaller(self, *rest):
        if self.active:
            self.active.menu_incsize(-2)
        else:
            EasyDialogs.Message("No active window?")

    def updatemenubar(self):
        changed = 0
        on = (self.active <> None)
        if on <> self.windowgroup_on:
            for m in self.windowgroup:
                m.enable(on)
            self.windowgroup_on = on
            changed = 1
        if on:
            # only if we have an edit menu
            on = self.active.have_selection()
            if on <> self.focusgroup_on:
                for m in self.focusgroup:
                    m.enable(on)
                self.focusgroup_on = on
                changed = 1
            on = self.active.can_paste()
            if on <> self.pastegroup_on:
                self.pasteitem.enable(on)
                self.pastegroup_on = on
                changed = 1
            on = self.active.can_undo()
            if on <> self.undo_label:
                if on:
                    self.undoitem.enable(1)
                    self.undoitem.settext(on)
                    self.undo_label = on
                else:
                    self.undoitem.settext("Nothing to undo")
                    self.undoitem.enable(0)
                changed = 1
            if self.updatefontmenus():
                changed = 1
        if changed:
            DrawMenuBar()

    def updatefontmenus(self):
        info = self.active.getruninfo()
        if info == self.ffs_values:
            return 0
        # Remove old checkmarks
        if self.ffs_values == ():
            self.ffs_values = (None, None, None)
        font, face, size = self.ffs_values
        if font <> None:
            fnum = self.fontnames.index(font)
            self.fontitems[fnum].check(0)
        if face <> None:
            for i in range(len(self.faceitems)):
                if face & (1<<i):
                    self.faceitems[i].check(0)
        if size <> None:
            for i in range(len(self.sizeitems)):
                if SIZES[i] == size:
                    self.sizeitems[i].check(0)

        self.ffs_values = info
        # Set new checkmarks
        font, face, size = self.ffs_values
        if font <> None:
            fnum = self.fontnames.index(font)
            self.fontitems[fnum].check(1)
        if face <> None:
            for i in range(len(self.faceitems)):
                if face & (1<<i):
                    self.faceitems[i].check(1)
        if size <> None:
            for i in range(len(self.sizeitems)):
                if SIZES[i] == size:
                    self.sizeitems[i].check(1)
        # Set outline/normal for sizes
        if font:
            exists = getfontsizes(font, SIZES)
            for i in range(len(self.sizeitems)):
                if exists[i]:
                    self.sizeitems[i].setstyle(0)
                else:
                    self.sizeitems[i].setstyle(8)

    #
    # Apple menu
    #

    def do_about(self, id, item, window, event):
        EasyDialogs.Message("A simple single-font text editor based on WASTE")

    #
    # File menu
    #

    def open(self, *args):
        self._open(0)

    def openfile(self, *args):
        self._open(1)

    def _open(self, askfile):
        if askfile:
            path = EasyDialogs.AskFileForOpen(typeList=('TEXT',))
            if not path:
                return
            name = os.path.split(path)[-1]
            try:
                fp = open(path, 'rb') # NOTE binary, we need cr as end-of-line
                data = fp.read()
                fp.close()
            except IOError, arg:
                EasyDialogs.Message("IOERROR: %r" % (arg,))
                return
        else:
            path = None
            name = "Untitled %d"%self.num
            data = ''
        w = WasteWindow(self)
        w.open(path, name, data)
        self.num = self.num + 1

    def insertfile(self, *args):
        if self.active:
            path = EasyDialogs.AskFileForOpen(typeList=('TEXT',))
            if not path:
                return
            try:
                fp = open(path, 'rb') # NOTE binary, we need cr as end-of-line
            except IOError, arg:
                EasyDialogs.Message("IOERROR: %r" % (args,))
                return
            self.active.menu_insert(fp)
        else:
            EasyDialogs.Message("No active window?")

    def inserthtml(self, *args):
        if self.active:
            path = EasyDialogs.AskFileForOpen(typeList=('TEXT',))
            if not path:
                return
            try:
                fp = open(path, 'r')
            except IOError, arg:
                EasyDialogs.Message("IOERROR: %r" % (arg,))
                return
            self.active.menu_insert_html(fp)
        else:
            EasyDialogs.Message("No active window?")


    def closewin(self, *args):
        if self.active:
            self.active.close()
        else:
            EasyDialogs.Message("No active window?")

    def save(self, *args):
        if self.active:
            self.active.menu_save()
        else:
            EasyDialogs.Message("No active window?")

    def saveas(self, *args):
        if self.active:
            self.active.menu_save_as()
        else:
            EasyDialogs.Message("No active window?")


    def quit(self, *args):
        for w in self._windows.values():
            w.close()
        if self._windows:
            return
        self._quit()

    #
    # Edit menu
    #

    def undo(self, *args):
        if self.active:
            self.active.menu_undo()
        else:
            EasyDialogs.Message("No active window?")

    def cut(self, *args):
        if self.active:
            self.active.menu_cut()
        else:
            EasyDialogs.Message("No active window?")

    def copy(self, *args):
        if self.active:
            self.active.menu_copy()
        else:
            EasyDialogs.Message("No active window?")

    def paste(self, *args):
        if self.active:
            self.active.menu_paste()
        else:
            EasyDialogs.Message("No active window?")

    def clear(self, *args):
        if self.active:
            self.active.menu_clear()
        else:
            EasyDialogs.Message("No active window?")

    #
    # Other stuff
    #

    def idle(self, event):
        if self.active:
            self.active.do_idle(event)
        else:
            Qd.SetCursor(Qd.GetQDGlobalsArrow())

    def newRuler(self, obj):
        """Insert a new ruler. Make it as wide as the window minus 2 pxls"""
        ted = obj.WEGetObjectOwner()
        l, t, r, b = ted.WEGetDestRect()
        return r-l, 4

    def drawRuler(self, (l, t, r, b), obj):
        y = (t+b)/2
        Qd.MoveTo(l+2, y)
        Qd.LineTo(r-2, y)
        return 0

class MyHTMLParser(htmllib.HTMLParser):

    def anchor_bgn(self, href, name, type):
        self.anchor = href
        if self.anchor:
            self.anchorlist.append(href)
            self.formatter.push_style('anchor')

    def anchor_end(self):
        if self.anchor:
            self.anchor = None
            self.formatter.pop_style()


def getfontnames():
    names = []
    for i in range(256):
        n = Fm.GetFontName(i)
        if n: names.append(n)
    return names

def getfontsizes(name, sizes):
    exist = []
    num = Fm.GetFNum(name)
    for sz in sizes:
        if Fm.RealFont(num, sz):
            exist.append(1)
        else:
            exist.append(0)
    return exist

def main():
    App = Wed()
    App.mainloop()

if __name__ == '__main__':
    main()



"""browsepict - Display all "PICT" resources found"""

import FrameWork
import EasyDialogs
from Carbon import Res
from Carbon import Qd
from Carbon import Win
from Carbon import Controls
from Carbon import List
import sys
import struct
import macresource

#
# Resource definitions
ID_MAIN=512
MAIN_LIST=1
MAIN_SHOW=2

# Where is the picture window?
LEFT=200
TOP=64

def main():
    macresource.need('DLOG', ID_MAIN, "PICTbrowse.rsrc")
    PICTbrowse()

class PICTbrowse(FrameWork.Application):
    def __init__(self):
        # First init menus, etc.
        FrameWork.Application.__init__(self)
        # Next create our dialog
        self.main_dialog = MyDialog(self)
        # Now open the dialog
        contents = self.findPICTresources()
        self.main_dialog.open(ID_MAIN, contents)
        # Finally, go into the event loop
        self.mainloop()

    def makeusermenus(self):
        self.filemenu = m = FrameWork.Menu(self.menubar, "File")
        self.quititem = FrameWork.MenuItem(m, "Quit", "Q", self.quit)

    def quit(self, *args):
        self._quit()

    def showPICT(self, resid):
        w = PICTwindow(self)
        w.open(resid)
        #EasyDialogs.Message('Show PICT %r' % (resid,))

    def findPICTresources(self):
        num = Res.CountResources('PICT')
        rv = []
        for i in range(1, num+1):
            Res.SetResLoad(0)
            try:
                r = Res.GetIndResource('PICT', i)
            finally:
                Res.SetResLoad(1)
            id, type, name = r.GetResInfo()
            rv.append((id, name))
        return rv

class PICTwindow(FrameWork.Window):
    def open(self, (resid, resname)):
        if not resname:
            resname = '#%r' % (resid,)
        self.resid = resid
        picture = Qd.GetPicture(self.resid)
        # Get rect for picture
        print repr(picture.data[:16])
        sz, t, l, b, r = struct.unpack('hhhhh', picture.data[:10])
        print 'pict:', t, l, b, r
        width = r-l
        height = b-t
        if width < 64: width = 64
        elif width > 480: width = 480
        if height < 64: height = 64
        elif height > 320: height = 320
        bounds = (LEFT, TOP, LEFT+width, TOP+height)
        print 'bounds:', bounds

        self.wid = Win.NewWindow(bounds, resname, 1, 0, -1, 1, 0)
        self.wid.SetWindowPic(picture)
        self.do_postopen()

class MyDialog(FrameWork.DialogWindow):
    "Main dialog window for PICTbrowse"

    def open(self, id, contents):
        self.id = id
        FrameWork.DialogWindow.open(self, ID_MAIN)
        self.dlg.SetDialogDefaultItem(MAIN_SHOW)
        self.contents = contents
        self.ctl = self.dlg.GetDialogItemAsControl(MAIN_LIST)
        h = self.ctl.GetControlData_Handle(Controls.kControlListBoxPart,
                        Controls.kControlListBoxListHandleTag)
        self.list = List.as_List(h)
        self.setlist()

    def setlist(self):
        self.list.LDelRow(0, 0)
        self.list.LSetDrawingMode(0)
        if self.contents:
            self.list.LAddRow(len(self.contents), 0)
            for i in range(len(self.contents)):
                v = repr(self.contents[i][0])
                if self.contents[i][1]:
                    v = v + '"' + self.contents[i][1] + '"'
                self.list.LSetCell(v, (0, i))
        self.list.LSetDrawingMode(1)
        self.list.LUpdate(self.wid.GetWindowPort().visRgn)

    def getselection(self):
        items = []
        point = (0,0)
        while 1:
            ok, point = self.list.LGetSelect(1, point)
            if not ok:
                break
            items.append(point[1])
            point = point[0], point[1]+1
        values = []
        for i in items:
            values.append(self.contents[i])
        return values

    def do_show(self, *args):
        selection = self.getselection()
        for resid in selection:
            self.parent.showPICT(resid)

    def do_close(self):
        self.close()

    def do_itemhit(self, item, event):
        if item == MAIN_SHOW:
            self.do_show()

main()



import FrameWork
import Wbase, Wcontrols
from Carbon import Ctl, Controls, Qd, Res
from types import *
import Wapplication

#_arrowright = Qd.GetPicture(472)
#_arrowdown = Qd.GetPicture(473)

_arrowright = Res.Resource(
        '\x00I\x00\x00\x00\x00\x00\n\x00\n\x11\x01\x01\x00\n\x00\x00\x00'
        '\x00\x00\n\x00\n\x90\x00\x02\x00\x00\x00\x00\x00\n\x00\n\x00\x00'
        '\x00\x00\x00\n\x00\n\x00\x00\x00\x00\x00\n\x00\n\x00\x00\x10\x00'
        '\x18\x00\x1c\x00\x1e\x00\x1f\x00\x1f\x00\x1e\x00\x1c\x00\x18\x00'
        '\x10\x00\xff')


class PopupControl(Wcontrols.ControlWidget):

    def __init__(self, possize, items=None, callback=None):
        if items is None:
            items = []
        procID = Controls.popupMenuProc|Controls.popupFixedWidth|Controls.useWFont
        Wcontrols.ControlWidget.__init__(self, possize, "", procID, callback, 0, 0, 0)
        self._items = items[:]

    def set(self, value):
        self._control.SetControlValue(value+1)

    def get(self):
        return self._control.GetControlValue() - 1

    def open(self):
        self.menu = menu = FrameWork.Menu(self._parentwindow.parent.menubar, 'Foo', -1)

        for i in range(len(self._items)):
            item = self._items[i]
            if type(item) == StringType:
                menuitemtext = object = item
            elif type(item) == TupleType and len(item) == 2:
                menuitemtext, object = item
                self._items[i] = object
            else:
                raise Wbase.WidgetsError, "illegal itemlist for popup menu"
            menuitem = FrameWork.MenuItem(menu, menuitemtext, None, None)

        self._calcbounds()
        self._control = Ctl.NewControl(self._parentwindow.wid,
                                        self._bounds,
                                        self._title,
                                        1,
                                        self._value,
                                        self.menu.id,
                                        self._max,
                                        self._procID,
                                        0)
        self.SetPort()
        self.enable(self._enabled)

    def close(self):
        self.menu.delete()
        return Wcontrols.ControlWidget.close(self)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        part = self._control.TrackControl(point, -1)
        if part:
            if self._callback:
                Wbase.CallbackCall(self._callback, 0, self._items[self.get()])


class PopupWidget(Wbase.ClickableWidget):

    """Simple title-less popup widget. Should be 16x16 pixels.
    Builds the menu items on the fly, good for dynamic popup menus."""

    def __init__(self, possize, items=None, callback=None):
        Wbase.ClickableWidget.__init__(self, possize)
        if items is None:
            items = []
        self._items = items
        self._itemsdict = {}
        self._callback = callback
        self._enabled = 1

    def close(self):
        Wbase.ClickableWidget.close(self)
        self._items = None
        self._itemsdict = {}

    def draw(self, visRgn = None):
        if self._visible:
            Qd.FrameRect(self._bounds)
            Qd.EraseRect(Qd.InsetRect(self._bounds, 1, 1))
            l, t, r, b = self._bounds
            l = l + 2
            t = t + 3
            pictframe = (l, t, l + 10, t + 10)
            Qd.DrawPicture(_arrowright, pictframe)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        self.menu = FrameWork.Menu(self._parentwindow.parent.menubar, 'Foo', -1)
        self._additems(self._items, self.menu)
        self.SetPort()
        l, t, r, b = self._bounds
        l, t = Qd.LocalToGlobal((l+1, t+1))
        Wbase.SetCursor("arrow")
        self.menu.menu.EnableMenuItem(0)
        reply = self.menu.menu.PopUpMenuSelect(t, l, 1)
        if reply:
            id = reply >> 16
            item = reply & 0xffff
            self._menu_callback(id, item)
        self._emptymenu()

    def set(self, items):
        self._items = items

    def get(self):
        return self._items

    def _additems(self, items, menu):
        from FrameWork import SubMenu, MenuItem
        menu_id = menu.id
        for item in items:
            if item == "-":
                menu.addseparator()
                continue
            elif type(item) == ListType:
                submenu = SubMenu(menu, item[0])
                self._additems(item[1:], submenu)
                continue
            elif type(item) == StringType:
                menuitemtext = object = item
            elif type(item) == TupleType and len(item) == 2:
                menuitemtext, object = item
            else:
                raise Wbase.WidgetsError, "illegal itemlist for popup menu"

            if menuitemtext[:1] == '\0':
                check = ord(menuitemtext[1])
                menuitemtext = menuitemtext[2:]
            else:
                check = 0
            menuitem = MenuItem(menu, menuitemtext, None, None)
            if check:
                menuitem.check(1)
            self._itemsdict[(menu_id, menuitem.item)] = object

    def _emptymenu(self):
        menus = self._parentwindow.parent.menubar.menus
        for id, item in self._itemsdict.keys():
            if menus.has_key(id):
                self.menu = menus[id]
                self.menu.delete()
        self._itemsdict = {}

    def _menu_callback(self, id, item):
        thing = self._itemsdict[(id, item)]
        if callable(thing):
            thing()
        elif self._callback:
            Wbase.CallbackCall(self._callback, 0, thing)


class PopupMenu(PopupWidget):

    """Simple title-less popup widget. Should be 16x16 pixels.
    Prebuilds the menu items, good for static (non changing) popup menus."""

    def open(self):
        self._calcbounds()
        self.menu = Wapplication.Menu(self._parentwindow.parent.menubar, 'Foo', -1)
        self._additems(self._items, self.menu)

    def close(self):
        self._emptymenu()
        Wbase.Widget.close(self)
        self._items = None
        self._itemsdict = {}
        self.menu = None

    def set(self, items):
        if self._itemsdict:
            self._emptymenu()
        self.menu = Wapplication.Menu(self._parentwindow.parent.menubar, 'Foo', -1)
        self._items = items
        self._additems(self._items, self.menu)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        self.SetPort()
        l, t, r, b = self._bounds
        l, t = Qd.LocalToGlobal((l+1, t+1))
        Wbase.SetCursor("arrow")
        self.menu.menu.EnableMenuItem(0)
        reply = self.menu.menu.PopUpMenuSelect(t, l, 1)
        if reply:
            id = reply >> 16
            item = reply & 0xffff
            self._menu_callback(id, item)


class FontMenu(PopupMenu):

    """A font popup menu."""

    menu = None

    def __init__(self, possize, callback):
        PopupMenu.__init__(self, possize)
        _makefontmenu()
        self._callback = callback
        self._enabled = 1

    def open(self):
        self._calcbounds()

    def close(self):
        del self._callback

    def set(self):
        raise Wbase.WidgetsError, "can't change font menu widget"

    def _menu_callback(self, id, item):
        fontname = self.menu.menu.GetMenuItemText(item)
        if self._callback:
            Wbase.CallbackCall(self._callback, 0, fontname)

    def click(self, point, modifiers):
        if not self._enabled:
            return
        _makefontmenu()
        return PopupMenu.click(self, point, modifiers)


def _makefontmenu():
    """helper for font menu"""
    if FontMenu.menu is not None:
        return
    import W
    FontMenu.menu = Wapplication.Menu(W.getapplication().menubar, 'Foo', -1)
    W.SetCursor('watch')
    for i in range(FontMenu.menu.menu.CountMenuItems(), 0, -1):
        FontMenu.menu.menu.DeleteMenuItem(i)
    FontMenu.menu.menu.AppendResMenu('FOND')


def _getfontlist():
    from Carbon import Res
    fontnames = []
    for i in range(1, Res.CountResources('FOND') + 1):
        r = Res.GetIndResource('FOND', i)
        fontnames.append(r.GetResInfo()[2])
    return fontnames



# Scan an Apple header file, generating a Python file of generator calls.

import sys
import os
from bgenlocations import TOOLBOXDIR, BGENDIR
sys.path.append(BGENDIR)
from scantools import Scanner

LONG = "Icons"
SHORT = "icn"
OBJECT = "NOTUSED"

def main():
    input = LONG + ".h"
    output = SHORT + "gen.py"
    defsoutput = TOOLBOXDIR + LONG + ".py"
    scanner = MyScanner(input, output, defsoutput)
    scanner.scan()
    scanner.close()
    print "=== Testing definitions output code ==="
    execfile(defsoutput, {}, {})
    print "=== Done scanning and generating, now importing the generated code... ==="
    exec "import " + SHORT + "support"
    print "=== Done.  It's up to you to compile it now! ==="

class MyScanner(Scanner):

    def destination(self, type, name, arglist):
        classname = "Function"
        listname = "functions"
        if arglist:
            t, n, m = arglist[0]
            # This is non-functional today
            if t == OBJECT and m == "InMode":
                classname = "Method"
                listname = "methods"
        return classname, listname

    def makeblacklistnames(self):
        return [
                "GetIconCacheData",
                "SetIconCacheData",
                # Constants with funny definitions
                "kSelectorAllHugeData",
                "kSelectorAllAvailableData",
                "svAllAvailableData",
                # Something in a comment accidentally seen as a const definition
                "err",
                # OS8 only
                'IconServicesTerminate',
                # Lazy, right now.
                "GetIconRefFromFileInfo"
                ]

    def makeblacklisttypes(self):
        return [
                "IconActionUPP",
                "IconGetterUPP",
                "CFragInitBlockPtr",
                "CGRect_ptr",
                ]

    def makerepairinstructions(self):
        return [
                ]

    def writeinitialdefs(self):
        self.defsfile.write("def FOUR_CHAR_CODE(x): return x\n")
        self.defsfile.write("from Carbon.Files import *\n")

if __name__ == "__main__":
    main()



"""macgen_info - Generate CodeWarrior project, config source, resource file"""
import EasyDialogs
import os
import sys
import macfs
import MacOS
import macostools
import macgen_rsrc
# Note: this depends on being frozen, or on sys.path already being
# modified by macmodulefinder.
import makeconfig

TEMPLATEDIR=os.path.join(sys.prefix, ':Mac:mwerks:projects:build.macfreeze')
PROJECT_TEMPLATE=os.path.join(TEMPLATEDIR, ':frozen.prj')
CONFIG_TEMPLATE=os.path.join(TEMPLATEDIR, ':templatefrozenconfig.c')
BUNDLE_TEMPLATE=os.path.join(TEMPLATEDIR, ':frozenbundle.rsrc')

def generate(output, module_dict, debug=0, with_ifdef=0):
    problems = 0
    output_created=0
    if not os.path.exists(output):
        print 'Creating project folder', output
        os.mkdir(output)
        output_created = 1
    # Resolve aliases, if needed
    try:
        fss, dummy1, dummy2 = macfs.ResolveAliasFile(output)
    except macfs.error:
        pass
    else:
        newname = fss.as_pathname()
        if newname != output:
            if debug:
                print 'Alias', output
                print 'Resolved to', newname
            output = newname
    # Construct the filenames
    dummy, outfile = os.path.split(output)
    build, ext = os.path.splitext(outfile)
    if build == 'build' and ext[0] == '.':
        # This is probably a good name for the project
        projname = ext[1:]
    else:
        projname = 'frozenapplet.prj'
    config_name = os.path.join(output, ':macfrozenconfig.c')
    project_name = os.path.join(output, ':' + projname + '.prj')
    resource_name = os.path.join(output, ':frozenmodules.rsrc')
    bundle_name = os.path.join(output, ':frozenbundle.rsrc')

    # Fill the output folder, if needed.
    if output_created:
        # Create the project, if needed
        if not os.path.exists(project_name):
            print 'Creating project', project_name
            if not os.path.exists(PROJECT_TEMPLATE):
                print '** No template CodeWarrior project found at', PROJECT_TEMPLATE
                print '   To generate standalone Python applications from source you need'
                print '   a full source distribution. Check http://www.cwi.nl/~jack/macpython.html'
                print '   for details.'
                problems = 1
            else:
                macostools.copy(PROJECT_TEMPLATE, project_name)
                print 'A template CodeWarrior project has been copied to', project_name
                print 'It is up to you to make the following changes:'
                print '- Change the output file name'
                print '- Change the search path, unless the folder is in the python home'
                print '- Add sourcefiles/libraries for any extension modules used'
                print '- Remove unused sources, to speed up the build process'
                print '- Remove unused resource files (like tcl/tk) for a smaller binary'
                problems = 1
                macostools.copy(BUNDLE_TEMPLATE, bundle_name)
                print 'A template bundle file has also been copied to', bundle_name
                print 'You may want to adapt signature, size resource, etc'


    # Create the resource file
    macgen_rsrc.generate(resource_name, module_dict, debug=debug)

    # Create the config.c file
    if not os.path.exists(CONFIG_TEMPLATE):
        print '** No template config.c found at', PROJECT_TEMPLATE
        print '   To generate standalone Python applications from source you need'
        print '   a full source distribution. Check http://www.cwi.nl/~jack/macpython.html'
        print '   for details.'
        problems = 1
    else:
        # Find elegible modules (builtins and dynamically loaded modules)
        c_modules = []
        for module in module_dict.keys():
            if module_dict[module].gettype() in ('builtin', 'dynamic'):
                # if the module is in a package we have no choice but
                # to put it at the toplevel in the frozen application.
                if '.' in module:
                    module = module.split('.')[-1]
                c_modules.append(module)
        ifp = open(CONFIG_TEMPLATE)
        ofp = open(config_name, 'w')
        makeconfig.makeconfig(ifp, ofp, c_modules, with_ifdef)
        ifp.close()
        ofp.close()
        MacOS.SetCreatorAndType(config_name, 'CWIE', 'TEXT')

    if warnings(module_dict):
        problems = 1
    return problems

def warnings(module_dict):
    problems = 0
    for name, module in module_dict.items():
        if module.gettype() not in ('builtin', 'module', 'dynamic', 'package'):
            problems = problems + 1
            print 'Warning: %s not included: %s %s'%(name, module.gettype(), module)
    return problems



""" Python Character Mapping Codec generated from '8859-11.TXT' with gencodec.py.

    Generated from mapping found in
    ftp://ftp.unicode.org/Public/MAPPINGS/ISO8859/8859-11.TXT

"""#"

import codecs

### Codec APIs

class Codec(codecs.Codec):

    def encode(self,input,errors='strict'):

        return codecs.charmap_encode(input,errors,encoding_map)

    def decode(self,input,errors='strict'):

        return codecs.charmap_decode(input,errors,decoding_map)

class StreamWriter(Codec,codecs.StreamWriter):
    pass

class StreamReader(Codec,codecs.StreamReader):
    pass

### encodings module API

def getregentry():

    return (Codec().encode,Codec().decode,StreamReader,StreamWriter)

### Decoding Map

decoding_map = codecs.make_identity_dict(range(256))
decoding_map.update({
        0x00a1: 0x0e01, #       THAI CHARACTER KO KAI
        0x00a2: 0x0e02, #       THAI CHARACTER KHO KHAI
        0x00a3: 0x0e03, #       THAI CHARACTER KHO KHUAT
        0x00a4: 0x0e04, #       THAI CHARACTER KHO KHWAI
        0x00a5: 0x0e05, #       THAI CHARACTER KHO KHON
        0x00a6: 0x0e06, #       THAI CHARACTER KHO RAKHANG
        0x00a7: 0x0e07, #       THAI CHARACTER NGO NGU
        0x00a8: 0x0e08, #       THAI CHARACTER CHO CHAN
        0x00a9: 0x0e09, #       THAI CHARACTER CHO CHING
        0x00aa: 0x0e0a, #       THAI CHARACTER CHO CHANG
        0x00ab: 0x0e0b, #       THAI CHARACTER SO SO
        0x00ac: 0x0e0c, #       THAI CHARACTER CHO CHOE
        0x00ad: 0x0e0d, #       THAI CHARACTER YO YING
        0x00ae: 0x0e0e, #       THAI CHARACTER DO CHADA
        0x00af: 0x0e0f, #       THAI CHARACTER TO PATAK
        0x00b0: 0x0e10, #       THAI CHARACTER THO THAN
        0x00b1: 0x0e11, #       THAI CHARACTER THO NANGMONTHO
        0x00b2: 0x0e12, #       THAI CHARACTER THO PHUTHAO
        0x00b3: 0x0e13, #       THAI CHARACTER NO NEN
        0x00b4: 0x0e14, #       THAI CHARACTER DO DEK
        0x00b5: 0x0e15, #       THAI CHARACTER TO TAO
        0x00b6: 0x0e16, #       THAI CHARACTER THO THUNG
        0x00b7: 0x0e17, #       THAI CHARACTER THO THAHAN
        0x00b8: 0x0e18, #       THAI CHARACTER THO THONG
        0x00b9: 0x0e19, #       THAI CHARACTER NO NU
        0x00ba: 0x0e1a, #       THAI CHARACTER BO BAIMAI
        0x00bb: 0x0e1b, #       THAI CHARACTER PO PLA
        0x00bc: 0x0e1c, #       THAI CHARACTER PHO PHUNG
        0x00bd: 0x0e1d, #       THAI CHARACTER FO FA
        0x00be: 0x0e1e, #       THAI CHARACTER PHO PHAN
        0x00bf: 0x0e1f, #       THAI CHARACTER FO FAN
        0x00c0: 0x0e20, #       THAI CHARACTER PHO SAMPHAO
        0x00c1: 0x0e21, #       THAI CHARACTER MO MA
        0x00c2: 0x0e22, #       THAI CHARACTER YO YAK
        0x00c3: 0x0e23, #       THAI CHARACTER RO RUA
        0x00c4: 0x0e24, #       THAI CHARACTER RU
        0x00c5: 0x0e25, #       THAI CHARACTER LO LING
        0x00c6: 0x0e26, #       THAI CHARACTER LU
        0x00c7: 0x0e27, #       THAI CHARACTER WO WAEN
        0x00c8: 0x0e28, #       THAI CHARACTER SO SALA
        0x00c9: 0x0e29, #       THAI CHARACTER SO RUSI
        0x00ca: 0x0e2a, #       THAI CHARACTER SO SUA
        0x00cb: 0x0e2b, #       THAI CHARACTER HO HIP
        0x00cc: 0x0e2c, #       THAI CHARACTER LO CHULA
        0x00cd: 0x0e2d, #       THAI CHARACTER O ANG
        0x00ce: 0x0e2e, #       THAI CHARACTER HO NOKHUK
        0x00cf: 0x0e2f, #       THAI CHARACTER PAIYANNOI
        0x00d0: 0x0e30, #       THAI CHARACTER SARA A
        0x00d1: 0x0e31, #       THAI CHARACTER MAI HAN-AKAT
        0x00d2: 0x0e32, #       THAI CHARACTER SARA AA
        0x00d3: 0x0e33, #       THAI CHARACTER SARA AM
        0x00d4: 0x0e34, #       THAI CHARACTER SARA I
        0x00d5: 0x0e35, #       THAI CHARACTER SARA II
        0x00d6: 0x0e36, #       THAI CHARACTER SARA UE
        0x00d7: 0x0e37, #       THAI CHARACTER SARA UEE
        0x00d8: 0x0e38, #       THAI CHARACTER SARA U
        0x00d9: 0x0e39, #       THAI CHARACTER SARA UU
        0x00da: 0x0e3a, #       THAI CHARACTER PHINTHU
        0x00db: None,
        0x00dc: None,
        0x00dd: None,
        0x00de: None,
        0x00df: 0x0e3f, #       THAI CURRENCY SYMBOL BAHT
        0x00e0: 0x0e40, #       THAI CHARACTER SARA E
        0x00e1: 0x0e41, #       THAI CHARACTER SARA AE
        0x00e2: 0x0e42, #       THAI CHARACTER SARA O
        0x00e3: 0x0e43, #       THAI CHARACTER SARA AI MAIMUAN
        0x00e4: 0x0e44, #       THAI CHARACTER SARA AI MAIMALAI
        0x00e5: 0x0e45, #       THAI CHARACTER LAKKHANGYAO
        0x00e6: 0x0e46, #       THAI CHARACTER MAIYAMOK
        0x00e7: 0x0e47, #       THAI CHARACTER MAITAIKHU
        0x00e8: 0x0e48, #       THAI CHARACTER MAI EK
        0x00e9: 0x0e49, #       THAI CHARACTER MAI THO
        0x00ea: 0x0e4a, #       THAI CHARACTER MAI TRI
        0x00eb: 0x0e4b, #       THAI CHARACTER MAI CHATTAWA
        0x00ec: 0x0e4c, #       THAI CHARACTER THANTHAKHAT
        0x00ed: 0x0e4d, #       THAI CHARACTER NIKHAHIT
        0x00ee: 0x0e4e, #       THAI CHARACTER YAMAKKAN
        0x00ef: 0x0e4f, #       THAI CHARACTER FONGMAN
        0x00f0: 0x0e50, #       THAI DIGIT ZERO
        0x00f1: 0x0e51, #       THAI DIGIT ONE
        0x00f2: 0x0e52, #       THAI DIGIT TWO
        0x00f3: 0x0e53, #       THAI DIGIT THREE
        0x00f4: 0x0e54, #       THAI DIGIT FOUR
        0x00f5: 0x0e55, #       THAI DIGIT FIVE
        0x00f6: 0x0e56, #       THAI DIGIT SIX
        0x00f7: 0x0e57, #       THAI DIGIT SEVEN
        0x00f8: 0x0e58, #       THAI DIGIT EIGHT
        0x00f9: 0x0e59, #       THAI DIGIT NINE
        0x00fa: 0x0e5a, #       THAI CHARACTER ANGKHANKHU
        0x00fb: 0x0e5b, #       THAI CHARACTER KHOMUT
        0x00fc: None,
        0x00fd: None,
        0x00fe: None,
        0x00ff: None,
})

### Encoding Map

encoding_map = codecs.make_encoding_map(decoding_map)
